/* *****************************************************************
 *
 *  $Id: f0m_back_end.h,v 1.1.1.1 2002/04/28 22:06:50 jml Exp $
 *
 *  Copyright (c) 1999 by Dep.Informatica of the FCT/UNL
 *  Time-stamp: <00/11/16 20:44:20 jml>
 *
 *  Joao Lourenco <Joao.Lourenco@di.fct.unl.pt>
 *
 *  Parallel and Distributed Processing Group
 *  Departamento de Informatica
 *  Faculdade de Ciencias e Tecnologia
 *  Universidade Nova de Lisboa
 *
 * -----------------------------------------------------------------
 *
 *  Description:
 *
 * -----------------------------------------------------------------
 */
#ifndef __F1_BACK_END_H
#define __F1_BACK_END_H

/* *****************************************************************
 *
 *  $Id: f0m_back_end.h,v 1.1.1.1 2002/04/28 22:06:50 jml Exp $
 *
 *  Copyright (c) 1999 by Dep.Informatica of the FCT/UNL
 *  Time-stamp: <99/06/24 19:28:26 jml>
 *
 *  Joao Lourenco <Joao.Lourenco@di.fct.unl.pt>
 *
 *  Parallel and Distributed Processing Group
 *  Departamento de Informatica
 *  Faculdade de Ciencias e Tecnologia
 *  Universidade Nova de Lisboa
 *
 * -----------------------------------------------------------------
 *
 *  Description: Implementation of the back-end of the services
 *				 provided by FIDDLE-1.
 *
 *		f0m_be_initialize ()		-- Initialize FIDDLE-1
 *		f0m_be_terminate ()		-- Terminate FIDDLE-1
 *		f0m_be_tids ()			-- List existing task IDs
 *		f0m_be_kill ()			-- Kill a target process
 *		f0m_be_attach ()			-- Attach to a running process
 *		f0m_be_detach ()			-- Detach from a running process
 *		f0m_be_symbol_file ()	-- Specify whre to find program symbols
 *		f0m_be_file ()			-- Load program and symbols into memory
 *		f0m_be_run ()			-- Run a program
 *		f0m_be_step ()			-- Executite until the next instruction
 *		f0m_be_next ()			-- Executite until the next instruction
 *		f0m_be_continue ()		-- Continue the execution
 *		f0m_be_finish ()			-- Execute until returning from
 *								   current stack frame
 *		f0m_be_break ()			-- Set a breakpoint
 *		f0m_be_delete ()			-- Delete a breakpoint
 *		f0m_be_set_variable ()	-- Change the contents of a variable
 *		f0m_be_call ()			-- Call a function in the current
 *								   context of the target process
 *		f0m_be_evaluate ()		-- Evaluate an expression in the current
 *								   context of the target process
 *		f0m_be_display ()		-- Evaluate and display  an expression 
 *								   every time the process execution stops
 *		f0m_be_undisplay ()		-- Undisplay an expression
 *		f0m_be_list ()			-- List the process source code
 *		f0m_be_info_locals ()	-- Get name and value of local variables
 *		f0m_be_info_args ()		-- Get name and value of function argument
 *		f0m_be_info_line ()		-- Get info about current line
 *		f0m_be_info_program ()	-- Get info about current program
 *		f0m_be_info_stack ()		-- Get info about current stack frames
 *		f0m_be_up ()				-- Go up in the stack frame list
 *		f0m_be_down ()			-- Go down in the stack frame list
 *		f0m_be_frame ()			-- Select a specific stack frame
 *		f0m_be_info_break ()		-- Get info about current breakpoint
 *		f0m_be_info_display ()	-- Get info about current display expressions
 *		f0m_be_thread ()			-- Select a thread
 *		f0m_be_info_threads ()	-- Get info about the existing threads
 *		f0m_be_tty ()			-- Set a TTY for future IO
 *		f0m_be_sendto ()			-- Send a command directly to the debugger
 *
 * -----------------------------------------------------------------
 */

#include <stdlib.h>
#include <assert.h>
#include <pthread.h>
#include "chain.h"
#include "f0s_lib.h"

/* #################################################################
 *		#DEFINES, TYPES and GLOBAL (EXTERN) VARIABLES
 * ################################################################# */



/* #################################################################
 *		PUBLIC functions
 * ################################################################# */
 
/* -----------------------------------------------------------------
 * Function:    int status = f0m_initialize (void)
 *
 * Description: Start the system thread
 *
 * Returns:     The status
 */
int f0m_be_initialize (void);


/* -----------------------------------------------------------------
 * Function:    int status = f0m_terminate (void)
 *
 * Description: Terminate the system thread
 *
 * Returns:     The status
 */
int f0m_be_terminate (void);


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f0m_tids (IN const char *options)
 *
 * Description: Get a list of all existing process TIDs.
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0m_be_tids (const char *options);


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f0m_kill (IN int tid)
 *
 * Description: Kill a process being debugger but let the CLD
 *				    available for a next usage
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0m_be_kill (int tid);


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f0m_attach (IN int tp_pid)
 *
 * Description: Find an unsused component-level debugger or start one
 *				if none was found.  Then attach the debugger to the
 *				process 'tp_pid'.
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0m_be_attach (int tp_pid);


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f0m_detach (IN int tid)
 *
 * Description: Detach a component-level debugger from the process
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0m_be_detach (int tid);


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f0m_symbol_file (IN int tid, 
 *											 IN const char *file)
 *
 * Description: Load the symbols
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0m_be_symbol_file (int tid, const char *file);


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f0m_file (IN const char *file)
 *
 * Description: Load a process image into memory
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0m_be_file (const char *file);


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f0m_run (IN int tid)
 *
 * Description: Start the execution of a process
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0m_be_run (int tid);


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f0m_step (IN int tid)
 *
 * Description: Continue the execution of the process until next
 *					instruction is reached.  Steps into the called
 *					functions.  Blocks the caller until the command is 
 *					 executed.
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0m_be_step (int tid);


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f0m_next (IN int tid)
 *
 * Description: Continue the execution of the process until next
 *					instruction is reached.  Steps over the called
 *					functions.  Blocks the caller until the command is 
 *					 executed.
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0m_be_next (int tid);


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f0m_continue (IN int tid)
 *
 * Description: Continue the execution of the process until a
 *				    breakpoint is reached or the execution finishes.
 *				    Blocks the caller until the command is executed.
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0m_be_continue (int tid);


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f0m_finish (IN int tid)
 *
 * Description: Continue the execution of the process until a
 *				    breakpoint is reached or the execution finishes.
 *				    Blocks the caller until the command is executed.
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0m_be_finish (int tid);


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f0m_break (IN int tid, IN const char *file,
 *									   IN int line, IN const char *function)
 *
 * Description: Set a breakpoint.  
 *              The arguments 'line' and 'function' are exclusive.
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0m_be_break (int tid, const char *file, int line, const char *function);


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f0m_delete (IN int tid, IN int bpid)
 *
 * Description: Remove a breakpoint.
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0m_be_delete (int tid, int bpid);


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f0m_set_variable (IN int tid,
 *									   IN const char *variable,
 *									   IN const char *expression)
 *
 * Description: Set 'variable' to 'expression'
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0m_be_set_variable (int tid, const char *variable,
						  const char *expression);


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f0m_evaluate (IN int tid,
 *									   IN const char *expression)
 *
 * Description: Evaluate an 'expression'
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0m_be_evaluate (int tid, const char *expression);


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f0m_display (IN int tid,
 *									     IN const char *expression)
 *
 * Description: Display 'expression' every time the program stops
 *				If 'expression == NULL' list list all expressions/values
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0m_be_display (int tid, const char *expression);


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f0m_undisplay (IN int tid, IN int dispid)
 *
 * Description: Remove a display expression
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0m_be_undisplay (int tid, int dispid);


/* -----------------------------------------------------------------
 * Function:	tkout_t *rep = f0m_call (IN int tid,
 *									   IN const char *expression)
 *
 * Description: Call a function and get it's return value
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0m_be_call (int tid, const char *expression);


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f0m_list (IN int tid, IN const char *file,
 *									   IN int start_line,
 *									   IN const char *function,
 *									   IN int n_lines)
 *
 * Description: List a source file.  
 *              The arguments 'start_line' and 'function' are exclusive.
 *              The arguments 'n_lines' is optional.
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0m_be_list (int tid, const char *file, int start_line,
				 const char *function, int n_lines);


/* -----------------------------------------------------------------
 * Function:	tkout_t *rep = f0m_tty (IN int tid,
 *									  IN const char *file)
 *
 * Description: Set the I/O tty for the process being debugged
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0m_be_tty (int tid, const char *file);


/* -----------------------------------------------------------------
 * Function:	tkout_t *rep = f0m_info_locals (IN int tid)
 *
 * Description: Get name and value of local variables
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0m_be_info_locals (int tid);


/* -----------------------------------------------------------------
 * Function:	tkout_t *rep = f0m_info_args (IN int tid)
 *
 * Description: Get name and value of local variables
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0m_be_info_args (int tid);


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f0m_info_line (IN int tid, IN const char *file,
 *									        IN int line,
 *									        IN const char *function)
 *
 * Description: Get information about a source line.  
 *              The arguments 'line' and 'function' are exclusive.
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0m_be_info_line (int tid, const char *file, int line,
					  const char *function);


/* -----------------------------------------------------------------
 * Function:	tkout_t *rep = f0m_info_program (IN int tid)
 *
 * Description: Get information about a running program.  
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0m_be_info_program (int tid);


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f0m_info_stack (IN int tid)
 *
 * Description: Get information about a running program.  
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0m_be_info_stack (int tid);


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f0m_up (IN int tid)
 *
 * Description: Go "up" on the stack
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0m_be_up (int tid);


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f0m_down (IN int tid)
 *
 * Description: Go "down" on the stack
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0m_be_down (int tid);


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f0m_frame (IN int tid, IN int frame)
 *
 * Description: Select a specific stack frame
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0m_be_frame (int tid, int frame);


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f0m_info_break (IN int tid)
 *
 * Description: Get information about a running program.  
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0m_be_info_break (int tid);


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f0m_info_display (IN int tid)
 *
 * Description: Get information about a running program.  
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0m_be_info_display (int tid);


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f0m_thread (IN int tid, IN int threadID)
 *
 * Description: Switch the current thread.
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0m_be_thread (int tid, int threadID);


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f0m_info_threads (IN int tid)
 *
 * Description: Get information about the existing threads.
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0m_be_info_threads (int tid);


/* -----------------------------------------------------------------
 * Function:    tkout_t *rep = f0m_sendto (IN int tid, IN const char *comand)
 *
 * Description: Send a command string directly to the debugger.
 *				'tkout' contains the output (NOT PARSED).
 *
 * Returns:     The 'Output Token'.
 */
tkout_t *f0m_be_sendto (int tid, const char *command);


#endif /* __F1_BACK_END_H */
