/* *****************************************************************
 *
 *  $Id: f0m_x_2_xml.c,v 1.2 2002/09/27 14:01:27 vrm Exp $
 *
 *  Copyright (c) 1999 by Dep.Informatica of the FCT/UNL
 *  Time-stamp: <02/09/27 12:28:09 jml>
 *
 *  Vitor Moreira <vrm@asc.di.fct.unl.pt>
 *
 *  Parallel and Distributed Processing Group
 *  Departamento de Informatica
 *  Faculdade de Ciencias e Tecnologia
 *  Universidade Nova de Lisboa
 *
 * -----------------------------------------------------------------
 *
 *  Description:
 *
 * -----------------------------------------------------------------
 */

#include "config.h"

#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <stdlib.h>

#include <expat.h>

#include <f0m_tkin.h>
#include <f0s_tkout.h>
#include <glob.h>

#include "f0m_x_2_xml.h"



#define MAX_BUFFER 20


PRIVATE char *
str_replace (const char *source, char *match, char *replace);

/*
 * MACROS
 * ------
 *
 * ADD_INT(tag)   adiciona "  <tag>...</tag>"
 *   STRUCTURE    nome da estrutura que inclui o campo que queremos
 *   tag          nome do campo que queremos
 *
 * ADD_STR(tag)   igual ao anterior, mas mostra uma STRING em vez dum INT
 *
 * ADD_EMPTY(tag) adiciona "<tag/>"
 *
 * ADD_START(tag) adiciona "<tag>"
 *
 * ADD_END(tag)   adiciona "</tag>"
 *
 */

#define ADD_INT(tag)											\
		long2string ((long)STRUCTURE tag, aux, MAX_BUFFER);		\
		STRING = mychar_append (STRING, "<" #tag ">");			\
		STRING = mychar_append (STRING, aux);					\
		STRING = mychar_append (STRING, "</" #tag ">")

#define ADD_STR(tag)											\
		if (STRUCTURE tag == NULL)								\
			ADD_EMPTY(tag);										\
		else													\
		{ char *aux2, *aux3;									\
			STRING = mychar_append (STRING, "<" #tag ">");		\
	        /* substitui o '<' por "&lt;" */					\
			aux2 = str_replace (STRUCTURE tag, "<", "&lt;");	\
	        /* substitui o '>' por "&gt;" */					\
			aux3 = str_replace (aux2, ">", "&gt;");				\
			STRING = mychar_append (STRING, aux3);				\
			STRING = mychar_append (STRING, "</" #tag ">");		\
			if (aux3 != aux2)									\
			free (aux3);										\
			if (aux2 != STRUCTURE tag)							\
			free (aux2);										\
	    }

#define ADD_EMPTY(tag)  STRING = mychar_append (STRING, "<" #tag "/>")

#define ADD_START(tag)  STRING = mychar_append (STRING, "<" #tag ">")

#define ADD_END(tag)    STRING = mychar_append (STRING, "</" #tag ">")




/*************************************** begin MYCHAR module ***/
#define MYCHAR_INCREMENT 4096

typedef struct mychar_s {
	char *str;
	int len;
	int max;
} mychar_t;
/***************************************** end MYCHAR module ***/

typedef struct
{
  int code;
  mychar_t *result;
} datawalk_t;


/*** begin: str_replace module **********************************/
/* THERE IS ANOTHER COPY OF THIS CODE IN FILE F0M_XML_2_X.C     */
/* source  = source string containing (or not) chars to replace */
/* match   = chars to be replaced                               */
/* replace = chars to replace match                             */
PRIVATE char *
str_replace (const char *source, char *match, char *replace)
{
  char *p,*p1;
  int size_match;
  int size_replace;
  int count_match;
  int size_final_string;
  char *final_string;


  assert (source!=NULL);
  assert (match!=NULL);
  assert (replace!=NULL);

  size_match = strlen (match);
  size_replace = strlen (replace);

  /* saber o numero de match em source */
  count_match=0;
  p=(char*)source;
  while ((p=strstr (p, match))!=NULL)
    {
      count_match++;
      p++;
    }
/*   printf ("count_match=%d\n", count_match); */

  /* verificar se existe realmente trabalho para fazer */
  if (count_match==0)
    return ((char*)source);

  /* calcular o tamanho da nova string */
  size_final_string = strlen (source);
  size_final_string -= count_match*size_match;
  size_final_string += count_match*size_replace;
  assert (size_final_string>0);

  /* alocar espao */
  final_string = (char*) calloc (size_final_string, sizeof (char));
  assert (final_string!=NULL);

  /* processar at ao ltimo token */
  p=(char*)source;
  while ((p1=strstr (p, match))!=NULL)
    {
      strncat (final_string, p, p1-p);
      if (size_replace>0)
	strcat (final_string, replace);
      p=p1+1;
    }
  /* processar o ltimo token */
  if (p1 < ((char*)source+strlen(source)))
    strcat (final_string, p);

  return final_string;
}
/*** end: str_replace module ***********************************/


/*************************************** begin MYCHAR module ***/
PRIVATE mychar_t *
mychar_append (mychar_t *s, const char *str)
{
  int l = strlen (str);

  while (s->len + l >= s->max) {
    /* More space needed */
    s->max += MYCHAR_INCREMENT;
    s->str = (char *)realloc (s->str, s->max);
    assert (s->str != NULL);
  }
  memcpy (s->str + s->len, str, l+1);
  s->len += l;
  return s;
}

PRIVATE char *
mychar_thestr (mychar_t *s)
{
  char *ss = s->str;
  
  free (s);
  return ss;
}

PRIVATE mychar_t *
mychar_new (void)
{
  mychar_t *s = malloc (sizeof (mychar_t)); // alterado por vrm
	
  assert (s != NULL);
  s->len = 0;
  s->max = MYCHAR_INCREMENT;
  s->str = calloc (s->max, sizeof (char));
  assert (s->str != NULL); // alterado por vrm
  return s;
}

/***************************************** end MYCHAR module ***/


/***************************************** begin LONG2STRING module ***/
/*
 * ***************************************************************************
 * * long2string()
 * *
 * * Description : Convert a "long" to a null-terminated string
 * *               (base = decimal)
 * * Arguments   : input = number to be converted
 * *               str = pointer to string (i.e. display buffer)
 * *               numdigits = maximum number of digits to display
 * * Returns     : Number of digits
 * ***************************************************************************
 * */
PRIVATE int
long2string (long input, char *str, int numdigits)
{
  register int digit, dr;
  char buffer[numdigits];
  int sign = 0;
  
  memset (buffer, '0', numdigits); // added by vrm

  /* Get sign */
  if (input < 0) {
    sign = 1;				/* 1 means negative number */
    input = -input;
  }
  
  /* Generate a reverse null-terminated string */
  for (digit = 0;  input > 0  &&  digit < numdigits;  digit++) {
    buffer[digit] = (input % 10) + '0';
    input = input / 10;
  }
  
  // alterado por vrm: quando sai do ciclo for, o digit e' incrementado (nao devia acontecer!) 
  if (digit>0)
    digit--; 
  
  /* Insert sign if needed */
  if (sign)
    str[0] = '-';
  
  /* Reverse the (reversed) string */
  for (dr = sign;  digit >= 0;  digit--, dr++)
    str[dr] = buffer[digit];
  str[dr] = 0;				/* null-terminate the string */
  
  /* Return the number of digits used */
  return dr;
}
/************************************ end LONG2STRING module ***/

/******************************* begin F0M_TKIN_2_XML module ***/
/*** tkin_data2xml ***/
PRIVATE void
tkin_data2xml (const tkin_t *tkin, datawalk_t *param)
{
  char aux[MAX_BUFFER];
  req_tids_t *d_tids;
  req_attach_t *d_attach;
  req_symbol_file_t *d_symbol_file;
  req_file_t *d_file;
  req_run_t *d_run;
  req_thread_t *d_thread;
  req_sendto_t *d_sendto;
  req_terminate_t *d_terminate;
  req_frame_t *d_frame;
  req_info_locals_t *d_info_locals;
  req_info_line_t *d_info_line;
  req_tty_t *d_tty;
  req_signal_t *d_signal;
  req_break_t *d_break;
  req_delete_t *d_delete;
  req_set_variable_t *d_set_variable;
  req_evaluate_t *d_evaluate;
  req_display_t *d_display;
  req_undisplay_t *d_undisplay;
  req_call_t *d_call;
  req_list_t *d_list;
  /* adicionado em 19/08/2002 */
  req_add_debugger_t *d_add_debugger;
  req_register_client_t	*d_register_client;
  req_clients_t *d_clients;

  /* estruturas que no contm nada */
  // req_info_program_t  req_finish_t  req_step_t  req_info_stack_t  req_info_threads_t
  // req_info_break_t    req_kill_t    req_next_t  req_continue_t
  // req_info_display_t  req_up_t      req_down_t  req_detach_t

#define STRING param->result

  switch (param->code)
    {
    case C_CLIENTS:
      d_clients = (req_clients_t*) &(tkin->data);
#define STRUCTURE d_clients->
      ADD_START(c_clients);
      ADD_STR(options);
      ADD_END(c_clients);
#undef STRUCTURE
      break;
    case C_ADD_DEBUGGER:
      d_add_debugger = (req_add_debugger_t*) &(tkin->data);
#define STRUCTURE d_add_debugger->
      ADD_START(c_add_debugger);
      ADD_STR(host);
      ADD_INT(dbg_pid);
      ADD_INT(tp_pid);
      ADD_STR(cmd_channel);
      ADD_STR(rep_channel);
      ADD_END(c_add_debugger);
#undef STRUCTURE
      break;
    case C_REGISTER_CLIENT:
       d_register_client = (req_register_client_t*) &(tkin->data);
#define STRUCTURE d_register_client->
      ADD_START(c_register_client);
      ADD_INT(pid);
      ADD_STR(filename);
      ADD_STR(hostaddr);
      ADD_END(c_register_client);
#undef STRUCTURE
     break;
    case C_TIDS:
      d_tids = (req_tids_t*) &(tkin->data);
#define STRUCTURE d_tids->
      ADD_START(c_tids);
      ADD_STR(options);
      ADD_END(c_tids);
#undef STRUCTURE
      break;
    case C_KILL:
      ADD_EMPTY(c_kill);
      break;
    case C_DETACH:
      ADD_EMPTY(c_detach);
      break;
    case C_STEP:
      ADD_EMPTY(c_step);
      break;
    case C_NEXT:
      ADD_EMPTY(c_next);
      break;
    case C_CONTINUE:
      ADD_EMPTY(c_continue);
      break;
    case C_FINISH:
      ADD_EMPTY(c_finish);
      break;
    case C_UP:
      ADD_EMPTY(c_up);
      break;
    case C_DOWN:
      ADD_EMPTY(c_down);
      break;
    case C_INFO_PROGRAM:
      ADD_EMPTY(c_info_program);
      break;
    case C_INFO_STACK:
      ADD_EMPTY(c_info_stack);
      break;
    case C_INFO_BREAK:
      ADD_EMPTY(c_info_break);
      break;
    case C_INFO_DISPLAY:
      ADD_EMPTY(c_info_display);
      break;
    case C_INFO_THREADS:
      ADD_EMPTY(c_info_threads);
      break;
    case C_ATTACH:
      d_attach = (req_attach_t*) &(tkin->data);
#define STRUCTURE d_attach->
      ADD_START(c_attach);
      ADD_INT(tp_pid);
      ADD_STR(host);
      ADD_END(c_attach);
#undef STRUCTURE
      break;
    case C_SYMBOL_FILE:
      d_symbol_file = (req_symbol_file_t*) &(tkin->data);
#define STRUCTURE d_symbol_file->
      ADD_START(c_symbol_file);
      ADD_STR(file);
      ADD_STR(host);
      ADD_END(c_symbol_file);
#undef STRUCTURE
      break;
    case C_FILE:
      d_file = (req_file_t*) &(tkin->data);
#define STRUCTURE d_file->
      ADD_START(c_file);
      ADD_STR(file);
      ADD_STR(host);
      ADD_END(c_file);
#undef STRUCTURE
      break;
    case C_RUN:
      d_run = (req_run_t*) &(tkin->data);
#define STRUCTURE d_run->
      ADD_START(c_run);
      ADD_STR(args);
      ADD_END(c_run);
#undef STRUCTURE
      break;
    case C_SIGNAL:
      d_signal = (req_signal_t*) &(tkin->data);
#define STRUCTURE d_signal->
      ADD_START(c_signal);
      ADD_INT(tid);
      ADD_INT(signum);
      ADD_END(c_signal);
#undef STRUCTURE
      break;
    case C_BREAK:
      d_break = (req_break_t*) &(tkin->data);
#define STRUCTURE d_break->
      ADD_START(c_break);
      ADD_STR(file);
      ADD_INT(line);
      ADD_STR(function);
      ADD_END(c_break);
#undef STRUCTURE
      break;
    case C_DELETE:
      d_delete = (req_delete_t*) &(tkin->data);
#define STRUCTURE d_delete->
      ADD_START(c_delete);
      ADD_INT(bpid);
      ADD_END(c_delete);
#undef STRUCTURE
      break;
    case C_SET_VARIABLE:
      d_set_variable = (req_set_variable_t*) &(tkin->data);
#define STRUCTURE d_set_variable->
      ADD_START(c_set_variable);
      ADD_STR(variable);
      ADD_STR(expression);
      ADD_END(c_set_variable);
#undef STRUCTURE
      break;
    case C_EVALUATE:
      d_evaluate = (req_evaluate_t*) &(tkin->data);
#define STRUCTURE d_evaluate->
      ADD_START(c_evaluate);
      ADD_STR(expression);
      ADD_END(c_evaluate);
#undef STRUCTURE
      break;
    case C_DISPLAY:
      d_display = (req_display_t*) &(tkin->data);
#define STRUCTURE d_display->
      ADD_START(c_display);
      ADD_STR(expression);
      ADD_END(c_display);
#undef STRUCTURE
      break;
    case C_UNDISPLAY:
      d_undisplay = (req_undisplay_t*) &(tkin->data);
#define STRUCTURE d_undisplay->
      ADD_START(c_undisplay);
      ADD_INT(dispid);
      ADD_END(c_undisplay);
#undef STRUCTURE
      break;
    case C_CALL:
      d_call = (req_call_t*) &(tkin->data);
#define STRUCTURE d_call->
      ADD_START(c_call);
      ADD_STR(expression);
      ADD_END(c_call);
#undef STRUCTURE
      break;
    case C_LIST:
      d_list = (req_list_t*) &(tkin->data);
#define STRUCTURE d_list->
      ADD_START(c_list);
      ADD_STR(file);
      ADD_INT(start_line);
      ADD_STR(function);
      ADD_INT(n_lines);
      ADD_END(c_list);
#undef STRUCTURE
      break;
    case C_TTY:
      d_tty = (req_tty_t*) &(tkin->data);
#define STRUCTURE d_tty->
      ADD_START(c_tty);
      ADD_STR(file);
      ADD_END(c_tty);
#undef STRUCTURE
      break;
    case C_FRAME:
      d_frame = (req_frame_t*) &(tkin->data);
#define STRUCTURE d_frame->
      ADD_START(c_frame);
      ADD_INT(frame);
      ADD_END(c_frame);
#undef STRUCTURE
      break;
/*     case C_INFO_LOCALS: */
/*       d_info_locals = (req_info_locals_t*) &(tkin->data); */
/* #define STRUCTURE d_info_locals-> */
/*       ADD_START(c_info_locals); */
/*       ADD_STR(var); */
/*       ADD_STR(value); */
/*       ADD_END(c_info_locals); */
/* #undef STRUCTURE */
/*       break; */
    case C_INFO_ARGS: /* de onde e' que eu fui buscar isto? */
    case C_INFO_LOCALS:
      ADD_EMPTY(c_info_locals_args);
      break;
    case C_INFO_LINE:
      d_info_line = (req_info_line_t*) &(tkin->data);
#define STRUCTURE d_info_line->
      ADD_START(c_info_line);
      ADD_STR(file);
      ADD_INT(line);
      ADD_STR(function);
      ADD_END(c_info_line);
#undef STRUCTURE
      break;
    case C_THREAD:
      d_thread = (req_thread_t*) &(tkin->data);
#define STRUCTURE d_thread->
      ADD_START(c_thread);
      ADD_INT(threadID);
      ADD_END(c_thread);
#undef STRUCTURE
      break;
    case C_SENDTO:
      d_sendto = (req_sendto_t*) &(tkin->data);
#define STRUCTURE d_sendto->
      ADD_START(c_sendto);
      ADD_STR(command);
      ADD_END(c_sendto);
#undef STRUCTURE
      break;
    case C_TERM_SERVER:
      d_terminate = (req_terminate_t*) &(tkin->data);
#define STRUCTURE d_terminate->
      ADD_START(c_terminate_server);
      ADD_INT(halt);
      ADD_END(c_terminate);
#undef STRUCTURE
      break;
    }
#undef STRING
}

/*** f0m_tkin_2_xml ***/
PUBLIC char *
f0m_tkin_2_xml (const tkin_t *tkin)
{
  char aux[MAX_BUFFER];
  datawalk_t param;

#define STRING param.result
#define STRUCTURE tkin->

  param.result = mychar_new();
  param.code = STRUCTURE code;

  ADD_START(?xml version="1.0" standalone="yes"?);
  ADD_START(tkin);


  ADD_INT(tid);
  ADD_INT(code);

/*   ADD_START (data); */
  tkin_data2xml (tkin, &param);
/*   ADD_END (data); */

  ADD_END(tkin);
  
#undef STRUCTURE

  return mychar_thestr(STRING);
}
/********************************* end F0M_TKIN_2_XML module ***/


/****************************** begin F0M_TKOUT_2_XML module ***/
/*** tkout_datawalk2xml ***/
PRIVATE int
tkout_datawalk2xml (void *data, void *param)
{
  char aux[MAX_BUFFER];
  datawalk_t *args = (datawalk_t*)param;

  rfile_t        *d_file;
  rattach_t      *d_attach;
  rbreak_t       *d_break;
  rdetach_t      *d_detach;
  rdisplay_t     *d_display;
  revaluate_t    *d_evaluate;
  rinfolocals_t  *d_infolocals;
  rfinish_t      *d_finish;
  rinfobreak_t   *d_infobreak;
  rinfoline_t    *d_infoline;
  rinfoprogram_t *d_infoprogram;
  rframe_t       *d_frame;
  rinfothreads_t *d_infothreads;
  rlist_t        *d_list;
  rrun_t         *d_run;
  rsendto_t      *d_sendto;
  rtids_t        *d_tids;
  pposition_t    *d_position;
  obphit_t       *d_bphit;
  osignal_t      *d_signal;
  onewthread_t   *d_newthread;
  othread_t      *d_thread;

  /* updated 19/08/2002 */
  radddebugger_t *d_adddebugger;
  rclients_t     *d_clients;

  
#undef STRING
#define STRING args->result

/*   ADD_START (node); */
  switch (args->code)
    {
    case R_ADD_DEBUGGER:
      d_adddebugger = (radddebugger_t*) data;
#define STRUCTURE d_adddebugger->
      ADD_START(r_add_debugger);
      ADD_INT (tid);
      ADD_END(r_add_debugger);
#undef STRUCTURE
      break;

    case R_CLIENTS:
      d_clients = (rclients_t*) data;
#define STRUCTURE d_clients->
      ADD_START(r_clients);
      ADD_INT (pid);
      ADD_STR (filename);
      ADD_STR (hostaddr);
      ADD_END(r_clients);
#undef STRUCTURE
      break;

    case R_FILE:
      d_file = (rfile_t*) data;
#define STRUCTURE d_file->
      ADD_START(r_file);
      ADD_INT (tid);
      ADD_END(r_file);
#undef STRUCTURE
      break;

    case R_ATTACH:
      d_attach = (rattach_t*) data;
#define STRUCTURE d_attach->
      ADD_START(r_attach);
      ADD_INT (tid);
      ADD_INT (tp_pid);
      ADD_END(r_attach);
#undef STRUCTURE
      break;

    case R_BREAK:
    case R_BREAK_NO_SOURCE:
      d_break = (rbreak_t*) data;
#define STRUCTURE d_break->
      if (args->code==R_BREAK)
	{ADD_START(r_break);}
      else
	{ADD_START(r_break_no_source);}

      ADD_INT (bpid);
      ADD_INT (address);
      ADD_STR (file);
      ADD_INT (line);

      if (args->code==R_BREAK)
	{ADD_END (r_break);}
      else
	{ADD_END (r_break_no_code);}
#undef STRUCTURE
      break;

    case R_DETACH:
      d_detach = (rdetach_t*) data;
#define STRUCTURE d_detach->
      ADD_START (r_detach);
      ADD_INT (tp_pid);
      ADD_END (r_detach);
#undef STRUCTURE
      break;

    case R_INFO_BREAK:
    case R_INFO_BREAK_WATCH_DATA:
    case R_INFO_WATCHPOINT:
      d_infobreak = (rinfobreak_t*) data;
#define STRUCTURE d_infobreak->

      if (args->code==R_INFO_BREAK) {ADD_START (r_info_break);}
      if (args->code==R_INFO_BREAK_WATCH_DATA) {ADD_START (r_info_break_watch_data);}
      if (args->code==R_INFO_WATCHPOINT) {ADD_START (r_info_watchpoint);}

      ADD_INT (bpid);
      ADD_STR (type);
      ADD_STR (disp);
      ADD_INT (enabled);
      ADD_INT (address);
      ADD_STR (function);
      ADD_STR (file);
      ADD_INT (line);
      ADD_INT (hit_counter);
      ADD_INT (ign_counter);
      ADD_STR (condition);

      if (args->code==R_INFO_BREAK) {ADD_END (r_info_break);}
      if (args->code==R_INFO_BREAK_WATCH_DATA) {ADD_END (r_info_break_watch_data);}
      if (args->code==R_INFO_WATCHPOINT) {ADD_END (r_info_watchpoint);}
#undef STRUCTURE
      break;

    case R_INFO_PROGRAM:
      d_infoprogram = (rinfoprogram_t*) data;
#define STRUCTURE d_infoprogram->
      ADD_START(r_info_program);
      ADD_INT (tp_pid);
      ADD_INT (address);
      ADD_STR (why);
      ADD_INT (bpid);
      ADD_STR (signal);
      ADD_END (r_info_program);
#undef STRUCTURE
      break;

    case R_INFO_THREADS:
      d_infothreads = (rinfothreads_t*) data;
#define STRUCTURE d_infothreads->
      ADD_START (r_info_threads);
      ADD_INT (active);
      ADD_INT (th_id);
      ADD_INT (tp_pid);
      ADD_INT (frame);
      ADD_STR (function);
      ADD_STR (arguments);
      ADD_STR (file);
      ADD_INT (line);
      ADD_END (r_info_threads);
#undef STRUCTURE
      break;

    case R_LIST:
      d_list = (rlist_t*) data;
#define STRUCTURE d_list->
      ADD_START (r_list);
      ADD_INT (line_number);
      ADD_STR (line_text);
      ADD_END (r_list);
#undef STRUCTURE
      break;

    case R_RUN:
      d_run = (rrun_t*) data;
#define STRUCTURE d_run->
      ADD_START (r_run);
      ADD_STR (file);
      ADD_END (r_run);
#undef STRUCTURE
      break;
      
    case R_SENDTO:
      d_sendto = (rsendto_t*) data;
#define STRUCTURE d_sendto->
      ADD_START (r_sendto);
      ADD_END (r_sendto);
#undef STRUCTURE
      break;

    case R_TIDS:
      d_tids = (rtids_t*) data;
#define STRUCTURE d_tids->
      ADD_START (r_tids);
      ADD_INT (gtid);
      ADD_INT (tp_pid);
      ADD_INT (dbg_pid);
      ADD_INT (attached);
      ADD_INT (ltid);
      ADD_STR (machine);
      ADD_END (r_tids);
#undef STRUCTURE
      break;

      /* uses RFRAME_T */
    case R_INFO_STACK:
      d_frame = (rframe_t*) data;
#define STRUCTURE d_frame->
      ADD_START (r_info_stack);
      ADD_INT (frame);
      ADD_INT (address);
      ADD_STR (function);
      ADD_STR (arguments);
      ADD_STR (file);
      ADD_INT (line);
      ADD_END (r_info_stack);
#undef STRUCTURE
      break;

    case O_DISPLAY:
    case R_DISPLAY_ONE_LINE:
    case R_DISPLAY_MANY_START:
    case R_DISPLAY_MANY_MIDDLE:
    case R_DISPLAY_MANY_END:
    case R_INFO_DISPLAY:
      d_display = (rdisplay_t*) data;
#define STRUCTURE d_display->

      if (args->code==O_DISPLAY)              {ADD_START (o_display);}
      if (args->code==R_DISPLAY_ONE_LINE)     {ADD_START (r_display_one_line);}
      if (args->code==R_DISPLAY_MANY_START)   {ADD_START (r_display_many_start);}
      if (args->code==R_DISPLAY_MANY_MIDDLE)  {ADD_START (r_display_many_middle);}
      if (args->code==R_DISPLAY_MANY_END)     {ADD_START (r_display_many_end);}
      if (args->code==R_INFO_DISPLAY)         {ADD_START (r_info_display);}

      ADD_INT (dispid);
      ADD_INT (enabled);
      ADD_STR (expression);
      ADD_STR (value);

      if (args->code==O_DISPLAY)              {ADD_END (o_display);}
      if (args->code==R_DISPLAY_ONE_LINE)     {ADD_END (r_display_one_line);}
      if (args->code==R_DISPLAY_MANY_START)   {ADD_END (r_display_many_start);}
      if (args->code==R_DISPLAY_MANY_MIDDLE)  {ADD_END (r_display_many_middle);}
      if (args->code==R_DISPLAY_MANY_END)     {ADD_END (r_display_many_end);}
      if (args->code==R_INFO_DISPLAY) {ADD_END (r_info_display);}
#undef STRUCTURE
      break;

    case R_EVALUATE:
    case R_EVALUATE_ONE_LINE:
    case R_EVALUATE_MANY_START:
    case R_EVALUATE_MANY_MIDDLE:
    case R_EVALUATE_MANY_END:
      d_evaluate = (revaluate_t*) data;
#define STRUCTURE d_evaluate->

      if (args->code==R_EVALUATE_ONE_LINE)     {ADD_START (r_evaluate_one_line);}
      if (args->code==R_EVALUATE_MANY_START)   {ADD_START (r_evaluate_many_start);}
      if (args->code==R_EVALUATE_MANY_MIDDLE)  {ADD_START (r_evaluate_many_middle);}
      if (args->code==R_EVALUATE_MANY_END)     {ADD_START (r_evaluate_many_end);}
      if (args->code==R_EVALUATE)              {ADD_START (r_evaluate);}

      ADD_INT (dbg_var);
      ADD_STR (value);

      if (args->code==R_EVALUATE_ONE_LINE)     {ADD_END (r_evaluate_one_line);}
      if (args->code==R_EVALUATE_MANY_START)   {ADD_END (r_evaluate_many_start);}
      if (args->code==R_EVALUATE_MANY_MIDDLE)  {ADD_END (r_evaluate_many_middle);}
      if (args->code==R_EVALUATE_MANY_END)     {ADD_END (r_evaluate_many_end);}
      if (args->code==R_EVALUATE)              {ADD_END (r_evaluate);}
#undef STRUCTURE
      break;

    case R_INFO_LOCALS:
    case R_INFO_LOCALS_ONE_LINE:
    case R_INFO_LOCALS_MANY_START:
    case R_INFO_LOCALS_MANY_MIDDLE:
    case R_INFO_LOCALS_MANY_END:
      d_infolocals = (rinfolocals_t*) data;
#define STRUCTURE d_infolocals->

      if (args->code==R_INFO_LOCALS_ONE_LINE)     {ADD_START (r_info_locals_one_line);}
      if (args->code==R_INFO_LOCALS_MANY_START)   {ADD_START (r_info_locals_many_start);}
      if (args->code==R_INFO_LOCALS_MANY_MIDDLE)  {ADD_START (r_info_locals_many_middle);}
      if (args->code==R_INFO_LOCALS_MANY_END)     {ADD_START (r_info_locals_many_end);}
      if (args->code==R_INFO_LOCALS)              {ADD_START (r_info_locals_args);}

      ADD_STR (var);
      ADD_STR (value);

      if (args->code==R_INFO_LOCALS_ONE_LINE)     {ADD_END (r_info_locals_one_line);}
      if (args->code==R_INFO_LOCALS_MANY_START)   {ADD_END (r_info_locals_many_start);}
      if (args->code==R_INFO_LOCALS_MANY_MIDDLE)  {ADD_END (r_info_locals_many_middle);}
      if (args->code==R_INFO_LOCALS_MANY_END)     {ADD_END (r_info_locals_many_end);}
      if (args->code==R_INFO_LOCALS)              {ADD_END (r_info_locals_args);}
#undef STRUCTURE
      break;

    case R_FINISH:
      d_finish = (rfinish_t*) data;
#define STRUCTURE d_finish->
      ADD_START (r_finish);
      ADD_INT (s_address);
      ADD_STR (s_function);
      ADD_STR (s_args);
      ADD_STR (s_file);
      ADD_INT (s_line);
      ADD_INT (dbg_var);
      ADD_STR (value);
      ADD_END (r_finish);
#undef STRUCTURE
      break;

    case R_INFO_LINE:
    case R_INFO_LINE_NO_CODE:
      d_infoline = (rinfoline_t*) data;
#define STRUCTURE d_infoline->

      if (args->code==R_INFO_LINE)
	{ADD_START (r_info_line);}
      else
	{ADD_START (r_info_line_no_code);}

      ADD_INT (line);
      ADD_STR (file);
      ADD_STR (start_function);
      ADD_INT (start_function_offset);
      ADD_STR (end_function);
      ADD_INT (end_function_offset);
      ADD_INT (start_address);
      ADD_INT (end_address);

      if (args->code==R_INFO_LINE)
	{ADD_END (r_info_line);}
      else
	{ADD_END (r_info_line_no_code);}
#undef STRUCTURE
      break;

    case R_CALL_VOID:
    case R_CONTINUE:
    case R_DELETE:
    case R_DISPLAY_EXPRESSION:
    case R_KILL:
    case R_NEXT:
    case R_SET_VARIABLE:
    case R_STEP:
    case R_SYMBOL_FILE:
    case R_THREAD:
    case R_TTY:
    case R_UNDISPLAY:
      break;

    case P_POSITION:
      d_position = (pposition_t*) data;
#define STRUCTURE d_position->
      ADD_START (p_position);
      ADD_STR (file);
      ADD_INT (line);
      ADD_INT (pos);
      ADD_STR (where);
      ADD_INT (frame);
      ADD_END (p_position);
#undef STRUCTURE
      break;

    case O_BPHIT:
      d_bphit = (obphit_t*) data;
#define STRUCTURE d_bphit->
      ADD_START (o_bphit);
      ADD_INT (bpid);
      ADD_STR (function);
      ADD_STR (args);
      ADD_STR (file);
      ADD_INT (line);
      ADD_END (o_bphit);
#undef STRUCTURE
      break;

    case O_SIGNAL:
      d_signal = (osignal_t*) data;
#define STRUCTURE d_signal->
      ADD_START (o_signal);
      ADD_STR (name);
      ADD_END (o_signal);
#undef STRUCTURE
      break;

    case O_NEW_THREAD:
      d_newthread = (onewthread_t*) data;
#define STRUCTURE d_newthread->
      ADD_START (o_new_thread);
      ADD_INT (id);
      ADD_END (o_new_thread);
#undef STRUCTURE
      break;

    case O_SWITCH_TO_THREAD:
      d_thread = (othread_t*) data;
#define STRUCTURE d_thread->
      ADD_START (o_switch_to_thread);
      ADD_INT (pid);
      ADD_END (o_switch_to_thread);
#undef STRUCTURE
      break;
    }

/*   ADD_END (node); */
#undef STRING
  return 1;
}


/*** tkout_data2xml ***/
PRIVATE void
tkout_data2xml (data_t *data, datawalk_t *param)
{
  char aux[MAX_BUFFER];
  chain_node_t *result;


#define STRING param->result
#define STRUCTURE data->

  if (!STRUCTURE str)
    {ADD_EMPTY(str);}
  else
    {ADD_STR(str);}
  ADD_INT(code);

  ADD_START (chain);

  ADD_START (size);
  sprintf (aux, "%d", chain_length (data->data));
  STRING = mychar_append (STRING, aux);
  ADD_END (size);

  param->code = STRUCTURE code;
  result = chain_walk (data->data, tkout_datawalk2xml, param);
  assert (result==NULL);

  ADD_END (chain);

#undef STRING
#undef STRUCTURE
}


/*** tkout_oobwalk2xml ***/
PRIVATE int
tkout_oobwalk2xml (void *data, void *param)
{
  datawalk_t *args = (datawalk_t*)param;


#define STRING args->result
/*   ADD_START (data); */
  tkout_data2xml (data, args);
/*   ADD_END (data); */
#undef STRING
  return 1;
}


/*** f0m_tkout_2_xml ***/
PUBLIC char *
f0m_tkout_2_xml (const tkout_t *tkout)
{
  char aux[MAX_BUFFER];
  datawalk_t param;
  chain_node_t *result;


  /*** body of TKOUT ***/
  param.result = mychar_new();

#define STRING param.result
#define STRUCTURE tkout->

  ADD_START(?xml version="1.0" standalone="yes"?);
  ADD_START(tkout);

  ADD_INT(tid);
  ADD_STR(str);
  ADD_INT(status);

  /*** REP Section ***/
  if (!tkout->rep)
    { ADD_EMPTY(rep); }
  else
    {
      ADD_START(rep);
      tkout_data2xml (tkout->rep, &param);
      ADD_END(rep);
    }

  /*** OOB Section ***/
  if (chain_length (tkout->oob)==0)
    { 
      ADD_START(chain); 
      ADD_START (size);
      sprintf (aux, "0");
      STRING = mychar_append (STRING, aux);
      ADD_END (size);
      ADD_END(chain);
    }
  else
    {
      ADD_START (chain);
      
      ADD_START (size);
      sprintf (aux, "%d", chain_length (tkout->oob));
      STRING = mychar_append (STRING, aux);
      ADD_END (size);
      ADD_START (oob);
      result = chain_walk (tkout->oob, tkout_oobwalk2xml, &param);
      assert (result==NULL);
      ADD_END (oob);
      ADD_END (chain);
    }

  /*** POS Section ***/
  if (!tkout->pos)
    { ADD_EMPTY(pos); }
  else
    {
      ADD_START(pos);
      tkout_data2xml (tkout->pos, &param);
      ADD_END(pos);
    }

  /*** ERR Section ***/
  if (!tkout->err)
    { ADD_EMPTY(err); }
  else
    {
      ADD_START(err);
      tkout_data2xml (tkout->err, &param);
      ADD_END(err);
    }

  ADD_END(tkout);
  
  return mychar_thestr(STRING);
}
/******************************** end F0M_TKOUT_2_XML module ***/
