/* *****************************************************************
 *
 *  $Id: f0m_xml_2_x.c,v 1.1 2002/09/27 12:43:02 jml Exp $
 *
 *  Copyright (c) 1999 by Dep.Informatica of the FCT/UNL
 *  Time-stamp: <02/09/27 12:34:34 jml>
 *
 *  Vitor Moreira <vrm@asc.di.fct.unl.pt>
 *
 *  Parallel and Distributed Processing Group
 *  Departamento de Informatica
 *  Faculdade de Ciencias e Tecnologia
 *  Universidade Nova de Lisboa
 *
 * -----------------------------------------------------------------
 *
 *  Description:
 *
 * -----------------------------------------------------------------
 */

/*** INCLUDES **************************************************/
#include "config.h"

#include <stdio.h>
#include <string.h>
#include <assert.h>

#include <expat.h>

#include <f0m_tkin.h>
#include <f0s_tkout.h>
#include <glob.h>

#include "f0m_xml_2_x.h"


/*** DEFINES ***************************************************/
#define PRIVATE          static
#define LAST_EVENT_NONE  0
#define LAST_EVENT_START 1
#define LAST_EVENT_DATA  2
#define LAST_EVENT_END   3

#define STRLEN(s)	(s) == NULL ? 0 : strlen (s)
#define TEST(s)		(strncmp (pdata->tag, (s), STRLEN (s)) == 0)

#define CMP(s)		{int e = strncmp (pdata->tag, (s), STRLEN (s));		\
								assert (e == 0);}
#define NL          do { p = chain_next (stack, p);						\
                        pdata = (stack_t*)chain_node_data (p);			\
                        /* printf ("%s/", pdata->tag); \ */ \
                    } while (0)
#define EQ          {}
#define INT         (int)strtol (pdata->data, &pdata->data, 0)
#define X_II(i)     i = INT
#define X_AA(a)     a = (void *)INT
#define X_SS(s)     do { if (pdata->data==NULL)							\
                             (s) = NULL;								\
                         else											\
                             (s) = pdata->data;							\
                    } while (0)
#define X_S(n)      do { if (pdata->data==NULL)							\
                             { s[n].l = 0;								\
                               s[n].c = NULL; }							\
                         else											\
                             { s[n].c = strdup (pdata->data);			\
                               s[n].l = strlen (s[n].c); }				\
                    } while (0) 
#define X_I(n)      i[n] = INT
#define SKIP(c,n)   {}

/*** TYPEDEFS **************************************************/
typedef struct
{
  char *tag;    // etiqueta do elemento XML (file, tid, ...)
  char *data;   // valor do elemento XML (24,"fdshfj",...)
/*   int data_len; // espao ocupado (em chars) pelo campo data */
} stack_t;

typedef struct
{
  chain_t *stack;      // stack de armazenamento de informao (neste
                       //     caso,  uma lista)
  stack_t *pointer;    // armazenamento temporrio
  int     last_event;  // ltimo evento que ocorreu
} userdata_t;

typedef struct 
{
  chain_node_t *p;
  int          code;
  data_t       *data;
} ch_data_t;

/*** MACROS ****************************************************/
#define STACK_INIT(stack)		\
	stack = chain_create();		\
    assert (stack!=NULL);

#define STACK_INIT_POINTER(pointer)						\
    pointer = (stack_t*) calloc (1, sizeof (stack_t));	\
    assert (pointer!=NULL); 

#define STACK_SET_TAG(pointer,tagname,taglen)			\
    STACK_INIT_POINTER(pointer);						\
    pointer->tag = (taglen==0) ? strdup (tagname) : strndup (tagname, taglen);

#define STACK_SET_DATA(pointer,datainfo,datalen)		\
	if (pointer->data==NULL) {							\
        pointer->data = (datalen==0)					\
	        ? strdup (datainfo)							\
		    : strndup (datainfo, datalen);				\
    }													\
    else {												\
        pointer->data = (char*) realloc (pointer->data, \
					 strlen(pointer->data)+datalen+1);	\
        assert (pointer->data!=NULL);					\
        strncat (pointer->data, datainfo, datalen);		\
    }

#define STACK_ADD(userdata)										\
    chain_enqueue (userdata->stack, (void*)userdata->pointer);	\
    userdata->pointer = NULL;
                                   


/*** EXPAT_HANDLERS ********************************************/

/******************************************* TKOUT_START_HND ***/
PRIVATE void
tkout_start_hnd (void *data, const char *el, const char **attr)
{
  userdata_t *userdata = (userdata_t*) data;


  if (userdata->last_event == LAST_EVENT_START)
    {
      STACK_ADD(userdata);
    }
  
  STACK_SET_TAG (userdata->pointer, el, 0);
  userdata->last_event = LAST_EVENT_START;
}

/******************************************** TKOUT_DATA_HND ***/
PRIVATE void
tkout_data_hnd (void *data, const XML_Char *s, int len)
{
  userdata_t *userdata = (userdata_t*) data;
  char *aux;


  STACK_SET_DATA (userdata->pointer, s, len);
  userdata->last_event = LAST_EVENT_DATA;
}

/********************************************* TKOUT_END_HND ***/
PRIVATE void
tkout_end_hnd (void *data, const char *el)
{
  userdata_t *userdata = (userdata_t*) data;


  if (userdata->pointer!=NULL)
    STACK_ADD (userdata);
  userdata->last_event = LAST_EVENT_END;
}



PRIVATE int
print_stack (void *data, void *args)
{
  stack_t *stack = (stack_t*) data;

  printf ("0x%x  %s,%s\n", stack->tag, stack->tag, stack->data);
  return 1;
}

PRIVATE void *
tkout_extract_node (ch_data_t *zz, chain_t *stack)
{
  void *base;
  chain_node_t *p;
  stack_t *pdata;


  p = zz->p;
  pdata = (stack_t*) chain_node_data (p);
  
#define NEW(type) type *d; base = (type *)malloc (sizeof (*d)); d = base
  
  switch (zz->code) {
        case R_TIDS:
        {   NEW(rtids_t);
                CMP ("r_tid"); NL;
                CMP ("gtid"); EQ; X_II (d->gtid); NL;
		CMP ("tp_pid"); EQ; X_II (d->tp_pid); NL;
		CMP ("dbg_pid"); EQ; X_II (d->dbg_pid); NL;
		CMP ("attached"); EQ; X_II (d->attached); NL;
		CMP ("ltid"); EQ; X_II (d->ltid); NL;
		CMP ("machine"); EQ; X_SS(d->machine); NL;
/* 		NL; */
		break;
	}
	case R_CLIENTS:
	{   NEW(rclients_t);
		CMP ("r_clients"); NL;
		CMP ("pid"); EQ; X_II (d->pid); NL;
		CMP ("filename"); EQ; X_SS(d->filename); NL;
		CMP ("hostaddr"); EQ; X_SS(d->hostaddr); NL;
/* 		NL; */
		break;
	}
	case R_KILL:
	{   /* NEW(rkill_t); */
		CMP ("r_kill"); NL;
		/* Empty */
		NL;
		break;
	}
	case R_ADD_DEBUGGER:
	{   NEW(radddebugger_t);
		CMP ("r_add_debugger"); NL;
		CMP ("tid"); EQ; X_II(d->tid); NL;
/* 		NL; */
		break;
	}
	case R_ATTACH:
	{   NEW(rattach_t);
		CMP ("r_attach"); NL;
		CMP ("tid"); EQ; X_II(d->tid); NL;
		CMP ("tp_pid"); EQ; X_II(d->tp_pid); NL;
/* 		NL; */
		break;
	}
	case R_DETACH:
	{   NEW(rdetach_t);
		CMP ("r_detach"); NL;
		CMP ("tp_pid"); EQ; X_II(d->tp_pid); NL;
/* 		NL; */
		break;
	}
	case R_SYMBOL_FILE:
	{   /* NEW(rsymbolfile_t); */
		CMP ("r_symbol_file"); NL;
		/* Empty */
/* 		CMP ("file"); EQ; X_SS(d->file); NL; */
		NL;
		break;
	}
	case R_FILE:
	{   NEW(rfile_t);
		CMP ("r_file"); NL;
		CMP ("tid"); EQ; X_II(d->tid); NL;
/* 		CMP ("file"); EQ; X_SS(d->file); NL; */
/* 		NL; */
		break;
	}
	case R_RUN:
	{   NEW(rrun_t);
		CMP ("r_run"); NL;
		CMP ("file"); EQ; X_SS(d->file); NL;
/* 		NL; */
		break;
	}
	case R_STEP:
	{   /* NEW(rstep_t); */
		CMP ("r_step"); NL;
		/* Empty */
/* 		NL; */
		break;
	}
	case R_NEXT:
	{   /* NEW(rnext_t); */
		CMP ("r_next"); NL;
		/* Empty */
/* 		NL; */
		break;
	}
	case R_CONTINUE:
	{   /* NEW(rcontinue_t); */
		CMP ("r_continue"); NL;
		/* Empty */
/* 		NL; */
		break;
	}
	case R_FINISH:
	{   NEW(rfinish_t);
		CMP ("r_finish"); NL;
		CMP ("s_address"); EQ; X_AA(d->s_address); NL;
		CMP ("s_function"); EQ; X_SS(d->s_function); NL;
		CMP ("s_args"); EQ; X_SS(d->s_args); NL;
		CMP ("s_file"); EQ; X_SS(d->s_file); NL;
		CMP ("s_line"); EQ; X_II(d->s_line); NL;
		CMP ("dbg_var"); EQ; X_II(d->dbg_var); NL;
		CMP ("value"); EQ; X_SS(d->value); NL;
/* 		NL; */
		break;
	}
	case R_CALL_VOID:
	{   /* NEW(rcallvoid_t); */
		CMP ("r_call_void"); NL;
		/* Empty */
/* 		NL; */
   		break;
	}
	case R_BREAK:
	{   NEW(rbreak_t);
		CMP ("r_break"); NL;
		CMP ("bpid"); EQ; X_II(d->bpid); NL;
		CMP ("address"); EQ; X_AA(d->address); NL;
		CMP ("file"); EQ; X_SS(d->file); NL;
		CMP ("line"); EQ; X_II(d->line); NL;
/* 		NL; */
		break;
	}
	case R_DELETE:
	{   /* NEW(rdelete_t); */
		CMP ("r_delete"); NL;
		/* Empty */
/* 		NL; */
		break;
	}
	case R_SET_VARIABLE:
	{   /* NEW(rsetvariable_t); */
		CMP ("r_set_variable"); NL;
		/* Empty */
/* 		NL; */
		break;
	}
	case R_EVALUATE:
	{   NEW(revaluate_t);
		CMP ("r_evaluate"); NL;
		CMP ("dbg_var"); EQ; X_II(d->dbg_var); NL;
		CMP ("value"); EQ; X_SS(d->value); NL;
/* 		NL; */
		break;
	}
	case R_DISPLAY_EXPRESSION:
	{   /* NEW(rdisplayexpression_t); */
		CMP ("r_display_expression"); NL;
		/* Empty */
/* 		NL; */
		break;
	}
	case R_UNDISPLAY:
	{   /* NEW(rundisplay_t); */
		CMP ("r_undisplay"); NL;
		/* Empty */
/* 		NL; */
		break;
	}
	case R_LIST:
	{   NEW(rlist_t);
		CMP ("r_list"); NL;
		CMP ("line_number"); EQ; X_II (d->line_number); NL;
		CMP ("line_text"); EQ; X_SS (d->line_text); NL;
/* 		NL; */
		break;
	}
	case R_INFO_LOCALS:
	{   NEW(rinfolocals_t);
		CMP ("r_info_locals_args"); NL;
		CMP ("var"); EQ; X_SS(d->var); NL;
		CMP ("value"); EQ; X_SS(d->value); NL;
/* 		NL; */
		break;
	}
	case R_INFO_LINE_NO_CODE:
	case R_INFO_LINE:
	{   NEW(rinfoline_t);
		CMP ("r_info_line"); NL;
		CMP ("line"); EQ; X_II(d->line); NL; 
		CMP ("file"); EQ; X_SS(d->file); NL; 
		CMP ("start_function"); EQ;
		X_SS(d->start_function); NL;
		CMP ("start_function_offset"); EQ;
		X_II(d->start_function_offset); NL;
		CMP ("end_function"); EQ;
		X_SS(d->end_function); NL;
		CMP ("end_function_offset"); EQ;
		X_II(d->end_function_offset); NL;
		CMP ("start_address"); EQ;
		X_AA(d->start_address); NL;
		CMP ("end_address"); EQ;
		X_AA(d->end_address); NL;
/* 		NL; */
		break;
	}
	case R_INFO_PROGRAM:
	{   NEW(rinfoprogram_t);
		CMP ("r_info_program"); NL;
		CMP ("tp_pid"); EQ; X_II(d->tp_pid); NL;
		CMP ("address"); EQ;
		X_AA(d->address); NL;
		CMP ("why"); EQ; X_SS(d->why); NL;
		CMP ("bpid"); EQ; X_II(d->bpid); NL;
		CMP ("signal"); EQ; X_SS(d->signal); NL;
/* 		NL; */
		break;
	}
	case R_INFO_STACK:
	{   NEW(rframe_t);
		CMP ("r_info_stack"); NL;
		CMP ("frame"); EQ; X_II (d->frame); NL;
		CMP ("address"); EQ; X_AA(d->address); NL;
		CMP ("function"); EQ; X_SS (d->function); NL;
		CMP ("arguments"); EQ; X_SS (d->arguments); NL;
		CMP ("file"); EQ; X_SS (d->file); NL;
		CMP ("line"); EQ; X_II (d->line); NL;
/* 		NL; */
		break;
	}
	case R_INFO_BREAK:
	{   NEW(rinfobreak_t);
		CMP ("r_info_break"); NL;
		CMP ("bpid"); EQ; X_II (d->bpid); NL;
		CMP ("type"); EQ; X_SS (d->type); NL;
		CMP ("disp"); EQ; X_SS (d->disp); NL;
		CMP ("enabled"); EQ; X_II (d->enabled); NL;
		CMP ("address"); EQ; X_AA (d->address); NL;
		CMP ("function"); EQ; X_SS (d->function); NL;
		CMP ("file"); EQ; X_SS (d->file); NL;
		CMP ("line"); EQ; X_II (d->line); NL;
		CMP ("hit_counter"); EQ; X_II (d->hit_counter); NL;
		CMP ("ign_counter"); EQ; X_II (d->ign_counter); NL;
		CMP ("condition"); EQ; X_SS (d->condition); NL;
/* 		NL; */
		break;
	}
	case R_INFO_DISPLAY:
	{   NEW(rdisplay_t);
		CMP ("r_info_display"); NL;
		CMP ("dispid"); EQ; X_II (d->dispid); NL;
		CMP ("enabled"); EQ; X_II (d->enabled); NL;
		CMP ("expression"); EQ; X_SS (d->expression); NL;
		CMP ("value"); EQ; X_SS (d->value); NL;
/* 		NL; */
		break;
	}
	case R_INFO_THREADS:
	{   NEW(rinfothreads_t);
		CMP ("r_info_threads"); NL;
		CMP ("active"); EQ; X_II (d->active); NL;
		CMP ("th_id"); EQ; X_II (d->th_id); NL;
		CMP ("tp_pid"); EQ; X_II (d->tp_pid); NL;
		CMP ("frame"); EQ; X_AA (d->frame); NL;
		CMP ("function"); EQ; X_SS (d->function); NL;
		CMP ("arguments"); EQ; X_SS (d->arguments); NL;
		CMP ("file"); EQ; X_SS (d->file); NL;
		CMP ("line"); EQ; X_II (d->line); NL;
/* 		NL; */
		break;
	}
	case R_TTY:
	{   /* NEW(rtty_t); */
		CMP ("r_tty"); NL;
		/* Empty */
/* 		NL; */
		break;
	}
	case R_THREAD:
	{   /* NEW(rthread_t); */
		CMP ("r_thread"); NL;
		/* Empty */
/* 		NL; */
		break;
	}
	case R_SENDTO:
	{   NEW(rsendto_t);
		CMP ("r_sendto"); NL;
		d->output = zz->data->str;
/* 		NL; */
   		break;
	}
	case R_REGISTER_CLIENT:
	{   NEW(rclients_t);
		CMP ("r_register_client"); NL;
		CMP ("pid"); EQ; X_II (d->pid); NL;
		CMP ("filename"); EQ; X_SS(d->filename); NL;
		CMP ("hostaddr"); EQ; X_SS(d->hostaddr); NL;
/* 		NL; */
		break;
	}
	case O_DISPLAY:
	{   NEW(odisplay_t);
		CMP ("o_display"); NL;
		CMP ("dispid"); EQ; X_II (d->dispid); NL;
		CMP ("enabled"); EQ; X_II (d->enabled); NL;
		CMP ("expression"); EQ; X_SS (d->expression); NL;
		CMP ("value"); EQ; X_SS (d->value); NL;
/* 		NL; */
		break;
	}
	case O_BPHIT:
	{   NEW(obphit_t);
		CMP ("o_bphit"); NL;
		CMP ("bpid"); EQ; X_II(d->bpid); NL;
		CMP ("function"); EQ; X_SS (d->function); NL;
		CMP ("args"); EQ; X_SS (d->args); NL;
		CMP ("file"); EQ; X_SS (d->file); NL;
		CMP ("line"); EQ;  X_II (d->line); NL;
/* 		NL; */
		break;
	}
	case O_SIGNAL:
	{   NEW(osignal_t);
		CMP ("o_signal"); NL;
		CMP ("name"); EQ; X_SS (d->name); NL;
/* 		NL; */
		break;
	}
	case O_NEW_THREAD:
	{   NEW(onewthread_t);
		CMP ("o_new_thread"); NL;
		CMP ("id"); EQ; X_II(d->id); NL;
/* 		NL; */
		break;
	}
	case O_SWITCH_TO_THREAD:
	{   NEW(othread_t);
		CMP ("thread"); NL;
		CMP ("pid"); EQ;  X_II (d->pid); NL;
/*  		NL; */
		break;
	}
	case P_POSITION:
	{   NEW(pposition_t);
		CMP ("p_position"); NL;
		CMP ("file"); EQ; X_SS (d->file); NL;
		CMP ("line"); EQ;  X_II (d->line); NL;
		CMP ("pos"); EQ;  X_II (d->pos); NL;
		CMP ("where"); EQ; X_SS (d->where); NL;
		CMP ("frame"); EQ; X_AA (d->frame); NL;
/* 		NL; */
		break;
	}
	default:
	  printf ("\n\
#################################################################\n\
%s:%d:tkout_extract_node()---Unexpected code = %d\n\
#################################################################\n",
                        __FILE__, __LINE__, zz->code);
		break;
	}
 	zz->p = p; 
	return base;
}



PRIVATE chain_t *
tkout_extract_chain (void *(*f)(), const char *var, ch_data_t *zz, chain_t *stack)
{
  int n_elems;
  void *elem;
  chain_t *ch;
  chain_node_t *p;
  stack_t *pdata;


  p = zz->p;
  pdata = (stack_t*) chain_node_data (p);

  ch = chain_create ();
  assert (ch != NULL);

  if (TEST ("chain"))
    {
      CMP ("chain"); SKIP ('[', 1);
      
      NL; CMP ("size");      // added by vrm
      X_II (n_elems);
      SKIP (':', 1);
      /*   CMP (var); */
      NL;
      while (n_elems--) {
	zz->p = p;
	elem = f (zz, stack); 
	p = zz->p; 
	assert (elem != NULL); 
	chain_enqueue (ch, elem); 
      }
/*       NL; */
      zz->p = p;
    }
  return ch;
}


/*********************************************** STACK_2_OOB ***/
PRIVATE void *
stack_2_oob (ch_data_t *zz, chain_t *stack)
{
  chain_node_t *p;
  stack_t *pdata;  // added by vrm
  data_t *oob = NULL;
  
  p = zz->p;
  pdata = (stack_t*) chain_node_data (p);  // added by vrm
/*   printf ("%s/", pdata->tag); */

  CMP ("oob");				/* Check it's a "oobdata_t" */
  NL;							/* Skip line */
  if (TEST ("str")) {  // modified by vrm
    oob = f0s_tkout_data_new ();
    
    /* The 'oobdata_t' was not empty */
    CMP ("str"); EQ; X_SS(oob->str);
    NL;
    
    CMP ("code"); EQ; X_II(oob->code);
    NL;
    
    zz->code = oob->code;
    zz->data = NULL;
    zz->p = p;
    oob->data = tkout_extract_chain (tkout_extract_node, "data", zz, stack);
  }
/*   NL; */
/*   zz->p = p; */
  return oob;
}



/********************************************** STACK_2_DATA ***/
PRIVATE data_t *
stack_2_data (const char *varname, chain_node_t **p1, chain_t *stack)
{
  chain_node_t *p;
  stack_t *pdata;
  data_t *var;
  ch_data_t zz;


  p = *p1;
  pdata = (stack_t*) chain_node_data (p);
  //  printf ("%s/", pdata->tag);


  CMP(varname);
  NL;

  var = NULL;

  if (TEST ("str")) {
    var = f0s_tkout_data_new ();
    CMP ("str");  EQ; X_SS(var->str);  NL;
    CMP ("code"); EQ; X_II(var->code); NL;

    zz.code = var->code;
    zz.p = p;
    zz.data = var;

    var->data = tkout_extract_chain (tkout_extract_node, "data", &zz, stack);
    p = zz.p;
  }
/*   else */
/*     { NL; } */
/*   //  NL; */
  *p1=p;
  return var;
}

/********************************************* STACK_2_TKOUT ***/
PRIVATE tkout_t *
stack_2_tkout (chain_t *stack)
{
  chain_node_t *p;
  stack_t *pdata;
  tkout_t *tk;
  ch_data_t zz;

  char *outp;
  int stat;
  int tid;


  assert (stack!=NULL);
  p = chain_first (stack);
  pdata = (stack_t*) chain_node_data (p);
/*   printf ("%s/", pdata->tag); */

  CMP ("tkout");
  NL;
  CMP ("tid");				/* Check next field it's "tid" */
  EQ;							/* Skip till number */
  X_II(tid);					/* Extract Task ID form string */
  NL;
  CMP ("str");				/* Check if output is present" */
  EQ;							/* Skip till string */
  X_SS(outp);					/* Get output string */
  NL;							/* Skip line */
  CMP ("status");				/* Check next field it's "code" */
  EQ;							/* Skip till number */
  X_II(stat);					/* Extract status form string */
  NL;

  tk = f0s_tkout_new (stat);
  tk->tid = tid;
  tk->str = outp;
  tk->rep = stack_2_data ("rep", &p, stack);

  zz.code = -1;
  zz.p = p;
  zz.data = NULL;
  tk->oob = tkout_extract_chain (stack_2_oob, "oob", &zz, stack);
  p = zz.p;
  tk->pos = stack_2_data ("pos", &p, stack);
/*   tk->err = stack_2_data ("err", &p, stack); */

/*  	char *p; */

/* 	zz.code = -1; */
/* 	zz.str = &p; */
/* 	zz.data = NULL; */
/*   tk->oob = tkout_extract_chain (str_2_oob, "oob", &zz);  */
/* 	p = *zz.str; */
/* 	tk->pos = str_2_data ("pos", &p); */
/* 	tk->err = str_2_data ("err", &p); */
/* 	NL; */
  return tk;
}

/*** begin: str_replace module **********************************/
/* THERE IS ANOTHER COPY OF THIS CODE IN FILE F0M_X_2_XML.C     */
/* source  = source string containing (or not) chars to replace */
/* match   = chars to be replaced                               */
/* replace = chars to replace match                             */
PRIVATE char *
str_replace (const char *source, char *match, char *replace)
{
  char *p,*p1;
  int size_match;
  int size_replace;
  int count_match;
  int size_final_string;
  char *final_string;


  assert (source!=NULL);
  assert (match!=NULL);
  assert (replace!=NULL);

  size_match = strlen (match);
  size_replace = strlen (replace);

  /* saber o numero de match em source */
  count_match=0;
  p=(char*)source;
  while ((p=strstr (p, match))!=NULL)
    {
      count_match++;
      p++;
    }
/*   printf ("count_match=%d\n", count_match); */

  /* verificar se existe realmente trabalho para fazer */
  if (count_match==0)
    return ((char*)source);

  /* calcular o tamanho da nova string */
  size_final_string = strlen (source);
  size_final_string -= count_match*size_match;
  size_final_string += count_match*size_replace;
  assert (size_final_string>0);

  /* alocar espao */
  final_string = (char*) calloc (size_final_string, sizeof (char));
  assert (final_string!=NULL);

  /* processar at ao ltimo token */
  p=(char*)source;
  while ((p1=strstr (p, match))!=NULL)
    {
      strncat (final_string, p, p1-p);
      if (size_replace>0)
	strcat (final_string, replace);
      p=p1+1;
    }
  /* processar o ltimo token */
  if (p1 < ((char*)source+strlen(source)))
    strcat (final_string, p);

  return final_string;
}
/*** end: str_replace module ***********************************/


/******************************************* F0M_XML_2_TKOUT ***/
PUBLIC tkout_t *
f0m_xml_2_tkout (const char *str)
{
  XML_Parser parser;
  userdata_t userdata;
  tkout_t *tkout;
  char *new_str;


  /* remover o \032 */
  new_str = str_replace (str, "\032", "");
  
  STACK_INIT (userdata.stack);
  userdata.last_event = LAST_EVENT_NONE;

  /* cria o parser */
  parser = XML_ParserCreate (NULL);
  assert (parser!=NULL);

  /* associa handlers ao parser */
  XML_SetElementHandler (parser, tkout_start_hnd, tkout_end_hnd);
  XML_SetCharacterDataHandler (parser, tkout_data_hnd);

  /* associa dados a serem passados entre handlers */
  XML_SetUserData (parser, &userdata);

  /* executa o parser */
  XML_Parse (parser, new_str, strlen(new_str), 1);

  /* destroi o parser */
  XML_ParserFree (parser);

/*   chain_walk (userdata.stack, print_stack, NULL); */

  /* converte a stack num tkout */
  tkout = stack_2_tkout (userdata.stack);

  /* liberta memria caso ela tenha sido alocada */
  if (new_str!=str)
    free (new_str);

  return tkout;
}


PUBLIC tkin_t *
stack_2_tkin (chain_t *stack)
{
  tkin_t *tk;
  int code;
  int tid;
  struct {
    int   l;
    char *c;
  } s[3];						/* For strings */
  int i[3];					/* For integers */
	
  chain_node_t *p;  // added by vrm
  stack_t* pdata;  // added by vrm

  p = chain_first (stack);
  pdata = (stack_t*) chain_node_data (p);

  CMP ("tkin");				/* Check it's a "tkin" */
  NL;							/* Skip line */
  CMP ("tid");				/* Check next field it's "code" */
  EQ;							/* Skip till number */
  tid = INT;					/* Extract code form string */
  assert (tid == -1  ||  tid > 0);
  NL;
  CMP ("code");				/* Check next field it's "code" */
  EQ;							/* Skip till number */
  code = INT;					/* Extract code form string */
  assert (code > 0);
  NL;
  switch (code) 
    {
    case C_TIDS:
      CMP ("c_tids"); NL;		/* Verify main field */
      CMP ("options"); EQ;	/* Verify sub-field name */
      X_S(0);
/*       NL; */
      tk = f0m_tkin_new (tid, C_TIDS, s[0].c);
      break;
    case C_CLIENTS:
      CMP ("c_clients"); NL;		/* Verify main field */
      CMP ("options"); EQ; X_S(0);/*  NL; */
     /*  NL; */
      tk = f0m_tkin_new (tid, C_CLIENTS, s[0].c);
      break;
    case C_KILL:
      CMP ("c_kill");/*  NL; */
     /*  NL; */
      tk = f0m_tkin_new (tid, C_KILL);
      break;
    case C_ADD_DEBUGGER:
      CMP ("c_add_debugger"); NL;
      CMP ("host"); EQ; X_S(0); NL;
      CMP("dbg_pid"); EQ; X_I(0); NL;
      CMP("tp_pid"); EQ; X_I(1); NL;
      CMP ("cmd_channel"); EQ; X_S(1); NL;
      CMP ("rep_channel"); EQ; X_S(2);/*  NL; */
    /*   NL; */
      tk = f0m_tkin_new (tid, C_ADD_DEBUGGER,
			 s[0].c, i[0], i[1], s[1].c, s[2].c);
      break;
    case C_ATTACH:
      CMP ("c_attach"); NL;
      CMP("tp_pid"); EQ; X_I(0); NL;
      CMP ("host"); EQ; X_S(0);/*  NL; */
     /*  NL; */
      tk = f0m_tkin_new (tid, C_ATTACH, i[0], s[0].c);
      break;
    case C_DETACH:
      CMP ("c_detach");/*  NL; */
     /*  NL; */
      tk = f0m_tkin_new (tid, C_DETACH);
      break;
    case C_SYMBOL_FILE:
      CMP ("c_symbol_file"); NL;
      CMP ("file"); EQ; X_S(0);/*  NL; */
     /*  NL; */
      tk = f0m_tkin_new (tid, C_SYMBOL_FILE, s[0].c);
      break;
    case C_FILE:
      CMP ("c_file"); NL;
      CMP ("file"); EQ; X_S(0); NL;
      CMP ("host"); EQ; X_S(1); /* NL; */
/*       NL; */
      tk = f0m_tkin_new (tid, C_FILE, s[0].c, s[1].c);
      break;
    case C_RUN:
      CMP ("c_run"); NL;
      CMP ("args"); EQ; X_S(0); /* NL; */
/*       NL; */
      tk = f0m_tkin_new (tid, C_RUN, s[0].c);
      break;
    case C_STEP:
      CMP ("c_step");/*  NL; */
/*       NL; */
      tk = f0m_tkin_new (tid, C_STEP);
      break;
    case C_NEXT:
      CMP ("c_next"); /* NL; */
/*       NL; */
      tk = f0m_tkin_new (tid, C_NEXT);
      break;
    case C_CONTINUE:
      CMP ("c_continue");/*  NL; */
/*       NL; */
      tk = f0m_tkin_new (tid, C_CONTINUE);
      break;
    case C_FINISH:
      CMP ("c_finish"); /* NL; */
/*       NL; */
      tk = f0m_tkin_new (tid, C_FINISH);
      break;
    case C_SIGNAL:
      CMP ("c_signal"); NL;
      CMP ("tid"); EQ; X_I(0); NL;
      CMP ("signum"); EQ; X_I(1); /* NL; */
     /*  NL; */
      tk = f0m_tkin_new (tid, C_SIGNAL, i[0], i[1]);
      break;
    case C_BREAK:
      CMP ("c_break"); NL;
      CMP ("file"); EQ; X_S(0); NL;
      CMP ("line"); EQ; X_I(0); NL;
      CMP ("function"); EQ; X_S(1); /* NL; */
/*       NL; */
      tk = f0m_tkin_new (tid, C_BREAK, s[0].c, i[0], s[1].c);
      break;
    case C_DELETE:
      CMP ("c_delete"); NL;
      CMP ("bpid"); EQ; X_I(0);/*  NL; */
     /*  NL; */
      tk = f0m_tkin_new (tid, C_DELETE, i[0]);
      break;
    case C_SET_VARIABLE:
      CMP ("c_set_variable"); NL;
      CMP ("variable"); EQ; X_S(0); NL;
      CMP ("expression"); EQ; X_S(1);/*  NL; */
     /*  NL; */
      tk = f0m_tkin_new (tid, C_SET_VARIABLE, s[0].c, s[1].c);
      break;
    case C_EVALUATE:
      CMP ("c_evaluate"); NL;
      CMP ("expression"); EQ; X_S(0);/*  NL; */
     /*  NL; */
      tk = f0m_tkin_new (tid, C_EVALUATE, s[0].c);
      break;
    case C_DISPLAY:
      CMP ("c_display"); NL;
      CMP ("expression"); EQ; X_S(0);/*  NL; */
     /*  NL; */
      tk = f0m_tkin_new (tid, C_DISPLAY, s[0].c);
      break;
    case C_UNDISPLAY:
      CMP ("c_undisplay"); NL;
      CMP ("dispid"); EQ; X_I(0);/*  NL; */
    /*   NL; */
      tk = f0m_tkin_new (tid, C_UNDISPLAY, i[0]);
      break;
    case C_CALL:
      CMP ("c_call"); NL;
      CMP ("expression"); EQ; X_S(0);/*  NL; */
     /*  NL; */
      tk = f0m_tkin_new (tid, C_CALL, s[0].c);
      break;
    case C_LIST:
      CMP ("c_list"); NL;
      CMP ("file"); EQ; X_S(0); NL;
      CMP ("start_line"); EQ; X_I(0); NL;
      CMP ("function"); EQ; X_S(1); NL;
      CMP ("n_lines"); EQ; X_I(1); /* NL; */
/*       NL; */
      tk = f0m_tkin_new (tid, C_LIST, s[0].c, i[0], s[1].c, i[1]);
      break;
    case C_TTY:
      CMP ("c_tty"); NL;
      CMP ("file"); EQ; X_S(0);/*  NL; */
    /*   NL; */
      tk = f0m_tkin_new (tid, C_TTY, s[0].c);
      break;
    case C_UP:
      CMP ("c_up");/*  NL; */
   /*    NL; */
      tk = f0m_tkin_new (tid, C_UP);
      break;
    case C_DOWN:
      CMP ("c_down");/*  NL; */
     /*  NL; */
      tk = f0m_tkin_new (tid, C_DOWN);
      break;
    case C_FRAME:
      CMP ("c_frame"); NL;
      CMP ("frame"); EQ; X_I(0);/*  NL; */
    /*   NL; */
      tk = f0m_tkin_new (tid, C_FRAME, i[0]);
      break;
    case C_INFO_LOCALS:
    case C_INFO_ARGS:
      CMP ("c_info_locals_args");/*  NL; */
    /*   NL; */
      tk = f0m_tkin_new (tid, code);
      break;
    case C_INFO_LINE:
      CMP ("c_info_line"); NL;
      CMP ("file"); EQ; X_S(0); NL;
      CMP ("line"); EQ; X_I(0); NL;
      CMP ("function"); EQ; X_S(1);/*  NL; */
    /*   NL; */
      tk = f0m_tkin_new (tid, C_INFO_LINE, s[0].c, i[0], s[1].c);
      break;
    case C_INFO_PROGRAM:
      CMP ("c_info_program"); /* NL; */
    /*   NL; */
      tk = f0m_tkin_new (tid, C_INFO_PROGRAM);
      break;
    case C_INFO_STACK:
      CMP ("c_info_stack");/*  NL; */
     /*  NL; */
      tk = f0m_tkin_new (tid, C_INFO_STACK);
      break;
    case C_INFO_BREAK:
      CMP ("c_info_break");/*  NL; */
    /*   NL; */
      tk = f0m_tkin_new (tid, C_INFO_BREAK);
      break;
    case C_INFO_DISPLAY:
      CMP ("c_info_display");/*  NL; */
     /*  NL; */
      tk = f0m_tkin_new (tid, C_INFO_DISPLAY);
      break;
    case C_THREAD:
      CMP ("c_thread");
      CMP ("threadID"); EQ; X_I(0);/*  NL; */
    /*   NL; */
      tk = f0m_tkin_new (tid, C_THREAD, i[0]);
      break;
    case C_INFO_THREADS:
      CMP ("c_info_threads");/*  NL; */
    /*   NL; */
      tk = f0m_tkin_new (tid, C_INFO_THREADS);
      break;
    case C_SENDTO:
      CMP ("c_sendto");
      NL;
      CMP ("command"); EQ; X_S(0);/*  NL; */
     /*  NL; */
      tk = f0m_tkin_new (tid, C_SENDTO, s[0].c);
      break;
    case C_TERM_SERVER:
      CMP ("c_terminate_server"); NL;
      CMP ("halt"); EQ; X_I(0);/*  NL; */
     /*  NL; */
      tk = f0m_tkin_new (tid, C_TERM_SERVER, i[0]);
      break;
    case C_REGISTER_CLIENT:
      CMP ("c_register_client"); NL;
      CMP ("pid"); EQ; X_I(0); NL;
      CMP ("filename"); EQ; X_S(0);/*  NL; */
     /*  NL; */
      tk = f0m_tkin_new (tid, C_REGISTER_CLIENT, i[0], s[0].c);
      break;
    default:
      printf ("%s:%d: Invalid code = %d\n",
	      __FILE__, __LINE__, tk->code);
      break;
    }
/*   NL; */
  return tk;
}


/******************************************** F0M_XML_2_TKIN ***/
PUBLIC tkin_t *
f0m_xml_2_tkin (const char *str)
{
  XML_Parser parser;
  userdata_t userdata;
  tkin_t *tkin;

  
  STACK_INIT (userdata.stack);
  userdata.last_event = LAST_EVENT_NONE;

  /* cria o parser */
  parser = XML_ParserCreate (NULL);
  assert (parser!=NULL);

  /* associa handlers ao parser */
  XML_SetElementHandler (parser, tkout_start_hnd, tkout_end_hnd);
  XML_SetCharacterDataHandler (parser, tkout_data_hnd);

  /* associa dados a serem passados entre handlers */
  XML_SetUserData (parser, &userdata);

  /* executa o parser */
  XML_Parse (parser, str, strlen(str), 1);

  /* destroi o parser */
  XML_ParserFree (parser);

  chain_walk (userdata.stack, print_stack, NULL);

  /* converte a stack num tkout */
  tkin = stack_2_tkin (userdata.stack);

  return tkin;
}

