/* *****************************************************************
 *
 *  $Id: f0s_dbg_io.c,v 1.1.1.1 2002/04/28 22:06:50 jml Exp $
 *
 *  Copyright (c) 1999 by Dep.Informatica of the FCT/UNL
 *  Time-stamp: <99/08/04 15:24:20 jml>
 *
 *  Joao Lourenco <Joao.Lourenco@di.fct.unl.pt>
 *
 *  Parallel and Distributed Processing Group
 *  Departamento de Informatica
 *  Faculdade de Ciencias e Tecnologia
 *  Universidade Nova de Lisboa
 *
 * -----------------------------------------------------------------
 *
 *  Description:
 *
 * -----------------------------------------------------------------
 */

#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include "f0s_getdelim.h"
#include "f0s_dbase.h"
#include "f0s_dbg_io.h"
#include "config.h"



static int count_lines (const char *command)
{
	int count = 0;
	char *pos;

	for (pos = strchr (command, '\n');
		 pos != NULL;
		 pos = strchr (pos + 1, '\n'))
		++count;
	return ++count;
}



/* -----------------------------------------------------------------
 * Function:    char *reply = f0s_send_receive (IN int tid,
 *											   IN const char *command)
 *
 * Description: Send a command string to the debugger 'tid', wait
 *				for its reply and return the output in an string
 *
 * Returns:     An string with the resulting output generated
 *				     by the debugger
 */
char *f0s_send_receive (int tid, const char *command)
{
	int count;
	char *reply = NULL;
	FILE *wfd = f0s_dbase_get_wfd (tid);
	FILE *rfd = f0s_dbase_get_rfd (tid);
	if (wfd == NULL  ||  rfd == NULL)
		return NULL; /* Invalid 'tid' */
	fprintf (wfd, "%s\n", command);
	count = count_lines (command);
	while (count--) {
		if (reply != NULL)
			free (reply);
		reply = get_delim (rfd, "(gdb) ");
	}
	return reply;
}

