/* *****************************************************************
 *
 *  $Id: f0s_tkout.c,v 1.7 2002/08/13 15:31:17 jml Exp $
 *
 *  Copyright (c) 1999 by Dep.Informatica of the FCT/UNL
 *  Time-stamp: <02/08/13 16:25:36 jml>
 *
 *  Joao Lourenco <Joao.Lourenco@di.fct.unl.pt>
 *
 *  Parallel and Distributed Processing Group
 *  Departamento de Informatica
 *  Faculdade de Ciencias e Tecnologia
 *  Universidade Nova de Lisboa
 *
 * -----------------------------------------------------------------
 *
 *  Description:
 *
 * -----------------------------------------------------------------
 */

#include <stdlib.h>
#include <assert.h>
#include "chain.h"
#include "msg.h"
#include "f0s_tkout.h"
#include "config.h"



#define OOB_DELETE -98765

#define DESTROY_STR(s)		if ((s) != NULL) free ((char *)(s))
#define DESTROY_CHAIN(chain, function, code)						\
			if ((chain) != NULL) {	         						\
				chain_prune ((chain), function, (void *)code);		\
				chain_destroy (chain);								\
			}


/* ================================================================= */
/* ================================================================= */
/* ================================================================= */
/* ================================================================= */


/* -----------------------------------------------------------------
 * Function:    int destroy_data (void *obj, void *args)
 *
 * Description: Free the contents of a 'chain_t'
 *				'obj' points to the object to free
 *				'(int) args' contains the type of object in 'obj'
 *
 * Returns:     A non-negative value
 */
static int destroy_data (void *obj, void *args)
{
	switch ((int) args) {
	case R_TIDS:
		break;
	case R_CLIENTS:
		break;
	case R_KILL:
		break;
	case R_ADD_DEBUGGER:
		break;
	case R_ATTACH:
		break;
	case R_DETACH:
		break;
	case R_SYMBOL_FILE:
/* 		DESTROY_STR (((rsymbolfile_t *)obj)->file); */
		break;
	case R_FILE:
/* 		DESTROY_STR (((rfile_t *)obj)->file); */
		break;
	case R_RUN:
		DESTROY_STR (((rrun_t *)obj)->file);
		break;
	case R_STEP:
		break;
	case R_NEXT:
		break;
	case R_CONTINUE:
		break;
	case R_FINISH:
		DESTROY_STR (((rfinish_t *)obj)->s_function);
		DESTROY_STR (((rfinish_t *)obj)->s_args);
		DESTROY_STR (((rfinish_t *)obj)->s_file);
		DESTROY_STR (((rfinish_t *)obj)->value);
		break;
	case R_CALL_VOID:
		break;
	case R_BREAK:
		DESTROY_STR (((rbreak_t *)obj)->file);
		break;
	case R_BREAK_NO_SOURCE:
		break;
	case R_DELETE:
		break;
	case R_SET_VARIABLE:
		break;
	case R_EVALUATE:
		DESTROY_STR (((revaluate_t *)obj)->value);
		break;
	case R_INFO_LOCALS:
		DESTROY_STR (((rinfolocals_t *)obj)->var);
		DESTROY_STR (((rinfolocals_t *)obj)->value);
		break;
	case R_DISPLAY_EXPRESSION:
		break;
	case R_UNDISPLAY:
		break;
	case R_TTY:
		break;
	case R_LIST:
		DESTROY_STR (((rlist_t *)obj)->line_text);
		break;
	case R_INFO_LINE:
		DESTROY_STR (((rinfoline_t *)obj)->file);
		DESTROY_STR (((rinfoline_t *)obj)->start_function);
		DESTROY_STR (((rinfoline_t *)obj)->end_function);
		break;
	case R_INFO_LINE_NO_CODE:
		DESTROY_STR (((rinfoline_t *)obj)->file);
		DESTROY_STR (((rinfoline_t *)obj)->start_function);
		DESTROY_STR (((rinfoline_t *)obj)->end_function);
		break;
	case R_INFO_PROGRAM:
		DESTROY_STR (((rinfoprogram_t *)obj)->why);
		DESTROY_STR (((rinfoprogram_t *)obj)->signal);
		break;
	case R_INFO_STACK:
		DESTROY_STR (((rframe_t *)obj)->function);
		DESTROY_STR (((rframe_t *)obj)->arguments);
		DESTROY_STR (((rframe_t *)obj)->file);
		break;
	case R_INFO_BREAK:
		DESTROY_STR (((rinfobreak_t *)obj)->type);
		DESTROY_STR (((rinfobreak_t *)obj)->disp);
		DESTROY_STR (((rinfobreak_t *)obj)->function);
		DESTROY_STR (((rinfobreak_t *)obj)->file);
		DESTROY_STR (((rinfobreak_t *)obj)->condition);
		break;
	case R_INFO_DISPLAY:
		DESTROY_STR (((rdisplay_t *)obj)->expression);
		DESTROY_STR (((rdisplay_t *)obj)->value);
		break;
	case R_INFO_THREADS:
		DESTROY_STR (((rinfothreads_t *)obj)->function);
		DESTROY_STR (((rinfothreads_t *)obj)->arguments);
		DESTROY_STR (((rinfothreads_t *)obj)->file);
		break;
	case R_THREAD:
		break;
	case R_SENDTO:
/* 		DESTROY_STR (((rsendto_t *)obj)->output); */
		break;
	case R_INFO_BREAK_WATCH_DATA:
		DESTROY_STR (((rinfobreak_t *)obj)->type);
		DESTROY_STR (((rinfobreak_t *)obj)->disp);
		DESTROY_STR (((rinfobreak_t *)obj)->function);
		DESTROY_STR (((rinfobreak_t *)obj)->file);
		DESTROY_STR (((rinfobreak_t *)obj)->condition);
		break;
	case R_INFO_WATCHPOINT:
		DESTROY_STR (((rinfobreak_t *)obj)->type);
		DESTROY_STR (((rinfobreak_t *)obj)->disp);
		DESTROY_STR (((rinfobreak_t *)obj)->function);
		DESTROY_STR (((rinfobreak_t *)obj)->file);
		DESTROY_STR (((rinfobreak_t *)obj)->condition);
		break;
		/* ===================================================== */
	case O_DISPLAY:
		DESTROY_STR (((odisplay_t *)obj)->expression);
		DESTROY_STR (((odisplay_t *)obj)->value);
		break;
	case O_BPHIT:
		DESTROY_STR (((obphit_t *)obj)->function);
		DESTROY_STR (((obphit_t *)obj)->args);
		DESTROY_STR (((obphit_t *)obj)->file);
		break;
	case O_SIGNAL:
		DESTROY_STR (((osignal_t *)obj)->name);
		break;
	case O_NEW_THREAD:
		break;
	case O_SWITCH_TO_THREAD:
		break;
		/* ===================================================== */
	case P_POSITION:
		DESTROY_STR (((pposition_t *)obj)->file);
		DESTROY_STR (((pposition_t *)obj)->where);
		break;
		/* ===================================================== */
	default:
		msg_pid_write ("\n\
#################################################################\n\
f0s_destroy_data()---Unexpected code = %d\n\
#################################################################\n",
                        (int)args);
	}
	return 1;
}




/* ================================================================= */
/* ================================================================= */
/* ================================================================= */
/* ================================================================= */

/* -----------------------------------------------------------------
 * Function:    data_t *data_s = f0s_tkout_data_new (void)
 *
 * Description: Create a new empty Data for the Output Token.
 *
 * Returns:     Pointer to the allocated memory area.
 */
data_t *f0s_tkout_data_new (void)
{
	data_t *r = calloc (1, sizeof (*r));
	if (r != NULL) {
		r->code = E_UNKNOWN_CMD_CODE;
		r->str = NULL;
		r->data = chain_create ();
		assert (r->data != NULL);
	}
	return r;
}


/* -----------------------------------------------------------------
 * Function:    int status = f0s_tkout_data_delete_2 (void *obj, void *args)
 *
 * Description: Delete a Data in the Output Token.
 *
 * Returns:     Status
 */
int f0s_tkout_data_delete_2 (void *obj, void *args)
{
	data_t *r = (data_t *)obj;
	if (r == NULL) return 0;	/* Empty */
	DESTROY_STR (r->str);
	DESTROY_CHAIN (r->data, destroy_data, r->code);
	return 1;
}


/* -----------------------------------------------------------------
 * Function:    int status = f0s_tkout_oob_delete (chain_t *oob)
 *
 * Description: Delete a Oob in the Output Token.
 *
 * Returns:     Status
 */
int f0s_tkout_oob_delete (chain_t *oob)
{
	if (oob == NULL) return 0;	/* Empty */
	DESTROY_CHAIN (oob, f0s_tkout_data_delete_2, NULL);
	return 0;
}




/* -----------------------------------------------------------------
 * Function:    tkout_t *tkout = f0s_tkout_new (void)
 *
 * Description: Create a new empty Output Token.
 *
 * Returns:     Pointer to the allocated token.
 */
tkout_t *f0s_tkout_new (code_t status)
{
	tkout_t *t = (tkout_t *) calloc (1, sizeof (*t));
/* 	t->reqid = -1; */				/* Invalid value for reqid. Should be
								   filed later */
	t->status = status;			/* The token status (given as argument) */
	t->str = NULL;
	t->rep = NULL;
	t->pos = NULL;
	t->oob = chain_create ();
	t->err = NULL;
	return t;
}



/* -----------------------------------------------------------------
 * Function:    int status = f0s_tkout_delete (tkout_t *t)
 *
 * Description: Free all the 'strings' in the token and then free the
 *                  token's memory
 *
 * Returns:     The status
 */
int f0s_tkout_delete (tkout_t *t)
{
	DESTROY_STR ((char *)t->str);
	f0s_tkout_data_delete (t->rep);
	f0s_tkout_data_delete (t->pos);
	f0s_tkout_data_delete (t->err);
	f0s_tkout_oob_delete (t->oob);
	free (t);
	return 0;
}
