/* *****************************************************************
 *
 *  $Id: f1m_servio.c,v 1.2 2002/09/27 12:43:02 jml Exp $
 *
 *  Copyright (c) 1999 by Dep.Informatica of the FCT/UNL
 *  Time-stamp: <02/09/16 18:56:25 jml>
 *
 *  Joao Lourenco <Joao.Lourenco@di.fct.unl.pt>
 *
 *  Parallel and Distributed Processing Group
 *  Departamento de Informatica
 *  Faculdade de Ciencias e Tecnologia
 *  Universidade Nova de Lisboa
 *
 * -----------------------------------------------------------------
 *
 *  Description:
 *
 * -----------------------------------------------------------------
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <assert.h>

#include "glob.h"

#include "f0s_tkout.h"

#include "f0m_tkin.h"
#include "f0m_xdr.h"

#include "config.h"

/* #################################################################
 *		#DEFINES, TYPES and GLOBAL (PRIVATE) VARIABLES
 * ################################################################# */

#ifndef NULL
#define NULL (void *)0
#endif

/* #define DEBUG_IO 1 */

/* #################################################################
 *		PRIVATE functions
 * ################################################################# */

/* -----------------------------------------------------------------
 * Function:    int myread (int fd, char *buffer, int size)
 *
 * Description: Read 'size' characters to buffer from 'fd'
 *
 * Returns:     Number of characters read
 */
PRIVATE int myread (int fd, char *buffer, int size)
{
	int total;
	int nread;
	
	for (total = 0;  total < size; total += nread) {
		nread = read (fd, buffer + total, size - total);
		if (nread == 0)
			break;
	}
	return total;
}



/* -----------------------------------------------------------------
 * Function:    int my_read_line (int fd, char *buffer, int max)
 *
 * Description: Read one line from 'fd' with 'max' characters at most
 *
 * Returns:     Number of bytes read
 *				-1 if EOF
 */
PRIVATE int my_read_line (int fd, char *buffer, int max)
{
	char *p = buffer;
	do {
		/* Read one char from STDIN */
		int nread = read (fd, p, 1);
		assert (nread == 1  ||  (nread == 0  &&  p == buffer));
		
		/* Check for EOF */
		if (nread == 0) return -1;
		
		/* Test for EOLine and continue if not */
	} while (*p++ != '\n');
	*p = '\0';					/* Terminate the string */
	return (p - buffer);
}


/* #################################################################
 *		PUBLIC functions
 * ################################################################# */



/* -----------------------------------------------------------------
 * Function:    tkout_t *f1m_read_from_server (int rfd)
 *
 * Description: Read a 'tkout_t' from a server
 *
 * Returns:     A pointer to the 'tkout_t' structure
 *				NULL on error
 */
tkout_t *f1m_read_from_server (int rfd)
{
	int err;
	char buffer[256];
	char *out_str;
	int out_len;
	tkout_t *tk;
	
	/* Pre-conditions */
	assert (rfd >= 0);

	/* Get the size of the reply */
	err = my_read_line (rfd, buffer, 256);
	assert (err > 0);
	out_len = -1;
	sscanf (buffer, "size = %d\n", &out_len);
	assert (out_len > 0);

	/* Get the reply in text form and convert it to a 'tkout_t' structure */
	out_str = alloca ((out_len+1) * sizeof (char));
	assert (out_str != NULL);
	err = myread (rfd, out_str, out_len);
	assert (err == out_len);
	out_str[out_len] = '\0';
	tk = f0m_xdr_2_tkout (out_str);
	assert (tk != NULL);
#ifdef DEBUG_IO
	puts ("---------- TKOUT ----------\n");
	puts (buffer);
	puts (out_str);
	puts ("---------------------------\n");
#endif
	
	return tk;
}



/* -----------------------------------------------------------------
 * Function:    int f1m_write_to_server (int wfd, const tkin_t *tk)
 *
 * Description: Write a 'tkin_t' to a server
 *
 * Returns:     The number of bytes written.
 *				-1 on error
 */
int f1m_write_to_server (int wfd, const tkin_t *tk)
{
	int err;
	char size[256];
	char *in_str;
	int in_len;
	
	/* Pre-conditions */
	assert (tk != NULL);
	assert (wfd >= 0);

	/* Convert 'tk' into text and send it to the server*/
	in_str = f0m_tkin_2_xdr (tk);
	assert (in_str != NULL);

	/* Send the size of the 'tk' and the 'tk' itself to the server */
	in_len = strlen (in_str);
	sprintf (size, "size = %d\n", in_len);
	err = write (wfd, size, strlen (size));
	assert (err > 0);
	err = write (wfd, in_str, strlen (in_str));
	assert (err > 0);
	free (in_str);
#ifdef DEBUG_IO
	puts ("---------- TKIN ----------\n");
	puts (size);
	puts (in_str);
	puts ("--------------------------\n");
#endif
	return err;
}
