/* ******************************************************************
 *
 *  Mytest.java
 *
 *  Copyright (c) 2001 by Dep.Informatica of the FCT/UNL
 *  Time-stamp: <02/08/23 15:52:45 rja>
 *
 *  Ricardo Anastacio [jml] <rja@asc.di.fct.unl.pt>
 *
 *  Parallel and Distributed Processing Group
 *  Departamento de Informatica
 *  Faculdade de Ciencias e Tecnologia
 *  Universidade Nova de Lisboa
 *
 * -----------------------------------------------------------------
 *
 *  Description:
 *
 * -----------------------------------------------------------------
 */

import java.io.*;
import java.util.StringTokenizer;
import java.util.Enumeration;
import java.util.NoSuchElementException;

import unl.fct.di.fiddle_j.*;

public class Mytest {
	 Fiddle_j f;

	 public Mytest() {
		  super();
		  try {
				f = new Fiddle_j();
		  } catch (FiddleNotInitializedException e) {
				System.err.println("Fiddle not Initialized:");
				System.err.println(e.getMessage());
				System.err.println("\nAborting execution!");
				System.exit(-1);
		  }
	 }

	 public String prompt() {
		  byte[] b = new byte[255];
		  int n;

		  System.out.print("\n\n [type help for... help!] --> ");
		  try {
				n = System.in.read(b);
				return new String(b, 0, n-1);
		  } catch (IOException ioe) {
				System.err.println("No data was read.\nIOException:"+ioe.getMessage());
				return "z"; //Unknown command...
		  }
	 }

	 public boolean parse(String str) {
		  StringTokenizer stk = new StringTokenizer(str);
		  String cmd = stk.nextToken();
		  Tkout_j tk;
		  boolean exit = false;

		  int tid;
		  String file;
		  String machine;
		  String token;
		  String expression;

		  try {
				switch(Character.toLowerCase(cmd.charAt(0))) {
				case 'a': //Attach
					 machine = stk.nextToken();
					 int tp_pid = Integer.parseInt(stk.nextToken());
					 tk = f.attach(machine, tp_pid);
					 System.out.println("\nEnd of attach, tkout contents" + tk);
					 break;
				case 'b': //Break
					 tid = Integer.parseInt(stk.nextToken());
					 if(stk.countTokens() > 1) // file specified
						  file = stk.nextToken();
					 else
						  file = null;
					 token = stk.nextToken();
					 try {  // O ultimo argumento pode ser o nmero de uma linha...
						  int line = Integer.parseInt(token);
						  tk = f.setBreakpoint(tid, file, line, null);
					 } catch(NumberFormatException nfe) { // ...ou o nome duma funo
						  tk = f.setBreakpoint(tid, file, -1, token);
					 }
					 System.out.println("\nEnd of break, tkout contents" + tk);
					 break;
				case 'c': //Call, Continue
					 try {
					 switch(Character.toLowerCase(cmd.charAt(1))) {
					 case 'a': //cAll
						  tid = Integer.parseInt(stk.nextToken());
						  expression = stk.nextToken();
						  tk = f.call(tid, expression);
						  System.out.println("\nEnd of call, tkout contents" + tk);
						  break;
					 case 'o': //cOntinue
						  tid = Integer.parseInt(stk.nextToken());
						  tk = f.continueExecution(tid);
						  System.out.println("\nEnd of continue, tkout contents" + tk);
						  break;
					 default:
						  System.out.println("Unknown command.");
						  break;
					 }
					 } catch (StringIndexOutOfBoundsException e) {
						  System.out.println("Ambiguous command. It can be: call or continue. Decide!");
					 }
					 break;
				case 'd': //Delete, Detach, Display, Down
					 try { switch(Character.toLowerCase(cmd.charAt(2))) {
					 case 'l': //deLete
						  tid = Integer.parseInt(stk.nextToken());
						  int bpid = Integer.parseInt(stk.nextToken());
						  tk = f.delete(tid, bpid);
						  System.out.println("\nEnd of delete, tkout contents" + tk);
						  break;
					 case 't': //deTach
						  tid = Integer.parseInt(stk.nextToken());
						  tk = f.detach(tid);
						  System.out.println("\nEnd of detach, tkout contents" + tk);
						  break;
					 case 's': //diSplay
						  tid = Integer.parseInt(stk.nextToken());
						  expression = stk.nextToken();
						  tk = f.display(tid, expression);
						  System.out.println("\nEnd of display, tkout contents" + tk);
						  break;
					 case 'w': //doWn
						  tid = Integer.parseInt(stk.nextToken());
						  tk = f.down(tid);
						  System.out.println("\nEnd of down, tkout contents" + tk);
						  break;
					 default:
						  System.out.println("Unknown command.");
						  break;
					 }
					 } catch (StringIndexOutOfBoundsException e) {
						  System.out.println("Ambiguous command. It can be: delete, detach, display or down. Decide!");
					 }
					 
					 break;
				case 'e': //Evaluate
					 tid = Integer.parseInt(stk.nextToken());
					 expression = stk.nextToken();
					 tk = f.evaluate(tid, expression);
					 System.out.println("\nEnd of evaluate, tkout contents" + tk);
					 break;
				case 'f': //File, Finish, Frame
					 try { switch(Character.toLowerCase(cmd.charAt(2))) {
					 case 'l': //fiLe
						  machine = stk.nextToken();
						  file = stk.nextToken();
						  tk = f.file(machine, file);
						  System.out.println("\nEnd of file, tkout contents" + tk);
						  break;
					 case 'n': //fiNish
						  tid = Integer.parseInt(stk.nextToken());
						  tk = f.finish(tid);
						  System.out.println("\nEnd of finish, tkout contents" + tk);
						  break;
					 case 'a': //frAme
						  tid = Integer.parseInt(stk.nextToken());
						  int frame = Integer.parseInt(stk.nextToken());
						  tk = f.frame(tid, frame);
						  System.out.println("\nEnd of frame, tkout contents" + tk);
						  break;
					 default:
						  System.out.println("Unknown command.");
						  break;
					 }
					 } catch (StringIndexOutOfBoundsException e) {
						  System.out.println("Ambiguous command. It can be: file, finish or frame. Decide!");
					 }

					 break;
				case 'h':
					 help();
					 break;
				case 'i': //Info_break, Info_display, Info_line, Info_program, Info_stack, Info_threads
					 try { switch(Character.toLowerCase(cmd.charAt(5))) {
					 case 'b': //info_Break
						  tid = Integer.parseInt(stk.nextToken());
						  tk = f.info_break(tid);
						  System.out.println("\nEnd of info_break, tkout contents" + tk);
						  break;
					 case 'd': //info_Display
						  tid = Integer.parseInt(stk.nextToken());
						  tk = f.info_display(tid);
						  System.out.println("\nEnd of info_display, tkout contents" + tk);
						  break;
					 case 'l': //info_Line
						  tid = Integer.parseInt(stk.nextToken());
						  if(stk.countTokens() > 1) // file specified
								file = stk.nextToken();
						  else
								file = null;
						  token = stk.nextToken();
						  try {  // O ultimo argumento pode ser o nmero de uma linha...
								int line = Integer.parseInt(token);
								tk = f.info_line(tid, file, line, null);
						  } catch(NumberFormatException nfe) { // ...ou o nome duma funo
								tk = f.info_line(tid, file, -1, token);
						  }
						  System.out.println("\nEnd of info_line, tkout contents" + tk);
						  break;
					 case 'p': //info_Program
						  tid = Integer.parseInt(stk.nextToken());
						  tk = f.info_program(tid);
						  System.out.println("\nEnd of info_program, tkout contents" + tk);
						  break;
					 case 's': //info_Stack
						  tid = Integer.parseInt(stk.nextToken());
						  tk = f.info_stack(tid);
						  System.out.println("\nEnd of info_stack, tkout contents" + tk);
						  break;
					 case 't': //info_Threads
						  tid = Integer.parseInt(stk.nextToken());
						  tk = f.info_threads(tid);
						  System.out.println("\nEnd of info_threads, tkout contents" + tk);
						  break;
					 default:
						  System.out.println("Unknown command.");
						  break;
					 }
					 } catch (StringIndexOutOfBoundsException e) {
						  System.out.println("Ambiguous command. It can be: info_break, info_display, info_line, info_program, info_stack or info_threads. Decide!");
					 }
					 
					 break;
				case 'k': //Kill
					 tid = Integer.parseInt(stk.nextToken());
					 tk = f.kill(tid);
					 System.out.println("\nEnd of kill, tkout contents" + tk);
					 break;
				case 'l': //List
					 tid = Integer.parseInt(stk.nextToken(": \t\n\r\f"));
					 if(str.indexOf(':') != -1)  //file foi especificado
						  file = stk.nextToken();
					 else
						  file = null;
					 token = stk.nextToken();
					 try {  // O ultimo argumento pode ser o nmero de uma linha...
						  int line = Integer.parseInt(token);
						  int nLines;
						  if(stk.countTokens() > 0) // nLines specified
								nLines = Integer.parseInt(token);
						  else
								nLines = -1;
						  tk = f.list(tid, file, line, null, nLines);
					 } catch(NumberFormatException nfe) { // ...ou o nome duma funo
						  int nLines;
						  if(stk.countTokens() > 0) // nLines specified
								nLines = Integer.parseInt(token);
						  else
								nLines = -1;
						  tk = f.list(tid, file, -1, token, nLines);
					 }
					 System.out.println("\nEnd of line, tkout contents" + tk);
					 break;
				case 'n': //Next
					 tid = Integer.parseInt(stk.nextToken());
					 tk = f.next(tid);
					 System.out.println("\nEnd of next, tkout contents" + tk);
					 break;
				case 'q':
					 exit = true;
					 break;
				case 'r': //Run
					 tid = Integer.parseInt(stk.nextToken());
					 String args;
					 if(stk.hasMoreTokens())
						  args = stk.nextToken();
					 else
						  args = null;
					 tk = f.run(tid, args);
					 System.out.println("\nEnd of run, tkout contents" + tk);
					 break;
				case 's': //Sendto, Set_variable, Signal, Step, Symbol_file
					 try { switch(Character.toLowerCase(cmd.charAt(2))) {
					 case 'n': //seNdto
						  tid = Integer.parseInt(stk.nextToken());
						  String command = stk.nextToken();
						  tk = f.sendTo(tid, command);
						  System.out.println("\nEnd of sendTo, tkout contents" + tk);
						  break;
					 case 't': //seT_variable
						  tid = Integer.parseInt(stk.nextToken());
						  String variable = stk.nextToken();
						  expression = stk.nextToken();
						  tk = f.set_variable(tid, variable, expression);
						  System.out.println("\nEnd of set_variable, tkout contents" + tk);
						  break;
					 case 'g': //siGnal
						  tid = Integer.parseInt(stk.nextToken());
						  int signum = Integer.parseInt(stk.nextToken());
						  tk = f.signal(tid, signum);
						  System.out.println("\nEnd of signal, tkout contents" + tk);
						  break;
					 case 'e': //stEp
						  tid = Integer.parseInt(stk.nextToken());
						  tk = f.step(tid);
						  System.out.println("\nEnd of step, tkout contents" + tk);
						  break;
					 default:
						  System.out.println("Unknown command.");
						  break;
					 }
					 } catch (StringIndexOutOfBoundsException e) {
						  System.out.println("Ambiguous command. It can be: sendto, set_variable, signal, step or symbol_file. Decide!");
					 }

					 break;
				case 't': //Thread, Tids
					 try { switch(Character.toLowerCase(cmd.charAt(1))) {
					 case 'h': //tHread
						  tid = Integer.parseInt(stk.nextToken());
						  int threadID = Integer.parseInt(stk.nextToken());
						  tk = f.thread(tid, threadID);
						  System.out.println("\nEnd of thread, tkout contents" + tk);
						  break;
					 case 'i': //tIds
						  tk = f.tids();
						  System.out.println("\nEnd of tids, tkout contents" + tk);
						  break;
					 default:
						  System.out.println("Unknown command.");
						  break;
					 }
					 } catch (StringIndexOutOfBoundsException e) {
						  System.out.println("Ambiguous command. It can be: thread or tids. Decide!");
					 }

					 break;
				case 'u': //Undisplay, Up
					 try { switch(Character.toLowerCase(cmd.charAt(1))) {
					 case 'n': //uNdisplay
						  tid = Integer.parseInt(stk.nextToken());
						  int dispID = Integer.parseInt(stk.nextToken());
						  tk = f.undisplay(tid, dispID);
						  System.out.println("\nEnd of undisplay, tkout contents" + tk);
						  break;
					 case 'p': //uP
						  tid = Integer.parseInt(stk.nextToken());
						  tk = f.up(tid);
						  System.out.println("\nEnd of up, tkout contents" + tk);
						  break;
					 default:
						  System.out.println("Unknown command.");
						  break;
					 }
					 } catch (StringIndexOutOfBoundsException e) {
						  System.out.println("Ambiguous command. It can be: undisplay or up. Decide!");
					 }

					 break;
				default:
					 System.out.println("Unknown command.");
					 break;
				}
 		  } catch (NoSuchElementException nsee) {
				System.err.println("Wrong number of arguments.\nTry help.");
		  } catch (FiddleException fe) {
				System.err.println("Error in Fiddle: "+fe.getMessage() );
				try {
					 System.err.println("                   code = "+fe.code.value+", "+fe.code.getMessage());
				} catch (UnknownCodeException uce) {}
		  } catch (DebuggerException de) { 
				System.err.println("Debugger couldn't execute command: "+de.getMessage() );
				try {
					 System.err.println("                   code = "+de.code.value+", "+de.code.getMessage());
				} catch (UnknownCodeException uce) {}
		  } catch (NumberFormatException nfe) {
				System.err.println("Wrong arguments.\nTry help.");
		  }


		  return exit;
	 }

	 public void help() {
		  System.out.println("\nattach <machine> <tp_pid>\nbreak  <tid> [file] <line|function>\ncall <tid> <expression>\ncontinue <tid>\ndelete <tid> <bpid>\ndetach <tid>\ndisplay <tid> <expression>\ndown <tid>\nevaluate <tid> <expression>\nfile <machine> <file>\nfinish <tid>\nframe <tid> <frame>\ninfo_break <tid>\ninfo_display <tid>\ninfo_line <tid> [file] <line|function>\ninfo_program <tid>\ninfo_stack <tid>\ninfo_threads <tid>\nkill <tid>\nlist <tid> [file:]<start_line|function> [n_lines]\nnext <tid>\nquit\nrun <tid> [args]\nsendto <tid> <cmd>\nset_variable <tid> <variable> <expression>\nsignal <tid> <signum>\nstep <tid>\nsymbol_file <tid> <machine> <file>\nthread <tid> <threadID>\ntids\nundisplay <tid> <dispid>\nup <tid>\n");
	 }


	 public static void main(String[] args) {
		  Mytest t = new Mytest();
		  boolean exit = false;
		  String result;
		  Tkout_j tk;
		  
		  while(!exit) {
				result = t.prompt();
				exit = t.parse(result);
		  }

		  System.out.println("\nSaindo do programa de teste.\nObrigado por testar o fiddle_j.\n");
	 }
	 
	 
	 
}
