/* *****************************************************************
 *
 *  $Id: do_err.c,v 1.1 2002/08/31 15:23:00 rja Exp $
 *
 *  Copyright (c) 2001 by Dep.Informatica of the FCT/UNL
 *  Time-stamp: <02/08/13 17:24:53 rja>
 *
 *  Ricardo Anastacio [jml] <rja@asc.di.fct.unl.pt>
 *
 *  Parallel and Distributed Processing Group
 *  Departamento de Informatica
 *  Faculdade de Ciencias e Tecnologia
 *  Universidade Nova de Lisboa
 *
 * -----------------------------------------------------------------
 *
 *  Description:
 *
 * -----------------------------------------------------------------
 */

#include "do_err.h"

/* #################################################################
 *		#DEFINES, TYPES and GLOBAL (PRIVATE) VARIABLES
 * ################################################################# */


/* #################################################################
 *		PRIVATE functions
 * ################################################################# */

#ifdef EBUG
PRIVATE void passo1(float f) {
  fprintf(stderr, "do_err->Passo %.2f.\n", f);
  fflush(stderr);
}

PRIVATE void passo2(float f) {
  fprintf(stderr, "do_status->Passo %.2f.\n", f);
  fflush(stderr);
}
#endif


/*
 * For every node in the data field of the err structure (data_t) this
 * function is aplyed, through the iteration of Chain_walk(...). For
 * every data in the node a corresponding JAVA image is created and
 * enqueued in the data field (chain_j) of the fiddle_j err object
 * (data_j).
 * (+) data, the data in the node (see chain_walk() );
 * (+) args, extra arguments. This is ignored!
 * (-) A success value. If negative the chain_walk function will stop
 * the traversing.
 */
PRIVATE int process_errData(void *data, void *args) {
  // Do nothing;
  // Return something positive!
  return 1;
}




/*
 * Constructs the signature string for a DebuggerException
 * constructor with a String and a Code_j arguments.
 * (+/-) buffer, the buffer where the string will be placed;
 * (-) the signature string returned.
 */
PRIVATE char *StrCode_sigV(char *buffer) {
  strcpy(buffer, "(Ljava/lang/String;");
  strcat(buffer, CODE_J_SIG);
  strcat(buffer, ")V");

  return buffer;
}




/*
 * Constructs the signature string for a FiddleException
 * constructor with a Code_j argument.
 * (+/-) buffer, the buffer where the string will be placed;
 * (-) the signature string returned.
 */
PRIVATE char *Code_sigV(char *buffer) {
  strcpy(buffer, "(L");
  strcat(buffer, CODE_J_SIG);
  strcat(buffer, ")V");

  return buffer;
}


/* #################################################################
 *		PUBLIC functions
 * ################################################################# */


PUBLIC jthrowable do_err(JNIEnv *env, tkout_t *tk) {
  jclass    codeCls;
  jclass    excCls;
  jmethodID mid;
  jstring   str;
  jobject   code;
  char      sig[MAX_STR];

#ifdef EBUG
  passo1(1);
#endif
  
// Set str to the C value of the tkout struct
  if(tk->err->str == NULL) tk->err->str = strdup("");
  str = (*env)->NewStringUTF(env, tk->err->str);
  
#ifdef EBUG
  passo1(2);
#endif
  
// Find the Code_j class and build a Code_j for the code field
  codeCls = (*env)->FindClass(env, CODE_J_SIG);
  if(codeCls == 0) fatalError("do_err:FindClass:Code_j\n", -1);
  mid = (*env)->GetMethodID(env, codeCls, "<init>", "(I)V");
  if(mid == 0) fatalError("do_err:GetMethodID:Code_j Constructor\n", -2);
  code = (*env)->NewObject(env, codeCls, mid, tk->err->code);
  
// Get the data field ID of the errObject and set it to the value in
// the C tkout_t structure.
// Ignore! Though, the data field does exist it contains absolutely nothing!
/*
  if(chain_walk(tk->err->data, process_errData, NULL) != NULL)
    fatalError("do_err:chain_walk:Error processing data.\n", -9);
*/
 
#ifdef EBUG
  passo1(3);
#endif
  
// Find the DebuggerException class and its constructor method ID
  excCls = (*env)->FindClass(env, DEBUGGEREXCEPTION_SIG);
  if(excCls == 0) fatalError("do_err:FindClass:DebuggerException\n", -33);
  mid = (*env)->GetMethodID(env, excCls, "<init>", StrCode_sigV(sig));
  if(mid == 0) fatalError("do_err:GetMethodID:DebuggerException Constructor\n", -2);

#ifdef EBUG
  passo1(4);
#endif
  
// return the new TkoutException  
  return (*env)->NewObject(env, excCls, mid, str, code);
  
}






PUBLIC jthrowable do_status(JNIEnv *env, tkout_t *tk) {
  jclass    codeCls;
  jclass    excCls;
  jmethodID mid;
  jobject   status;
  char      sig[MAX_STR];
  
#ifdef EBUG
  passo2(1);
#endif
  
// Find the Code_j class and build a Code_j for the status field
  codeCls = (*env)->FindClass(env, CODE_J_SIG);
  if(codeCls == 0) fatalError("do_status:FindClass:Code_j\n", -1);
  mid = (*env)->GetMethodID(env, codeCls, "<init>", "(I)V");
  if(mid == 0) fatalError("do_status:GetMethodID:Code_j Constructor\n", -2);
  status = (*env)->NewObject(env, codeCls, mid, tk->status);
	
#ifdef EBUG
  passo2(2);
#endif
  
// Find the FiddleException class and its constructor method ID
  excCls = (*env)->FindClass(env, FIDDLEEXCEPTION_SIG);
  if(excCls == 0) fatalError("do_status:FindClass:FiddleException\n", -3);
  mid = (*env)->GetMethodID(env, excCls, "<init>", Code_sigV(sig));
  if(mid == 0) fatalError("do_status:GetMethodID:FiddleException Constructor\n", -2);

#ifdef EBUG
  passo2(3);
#endif
  
// return the new TkoutException  
  return (*env)->NewObject(env, excCls, mid, status);
}



