/* *****************************************************************
 *
 *  $Id: do_pos.c,v 1.1 2002/08/31 15:23:00 rja Exp $
 *
 *  Copyright (c) 2001 by Dep.Informatica of the FCT/UNL
 *  Time-stamp: <02/08/20 12:39:15 rja>
 *
 *  Ricardo Anastacio [jml] <rja@asc.di.fct.unl.pt>
 *
 *  Parallel and Distributed Processing Group
 *  Departamento de Informatica
 *  Faculdade de Ciencias e Tecnologia
 *  Universidade Nova de Lisboa
 *
 * -----------------------------------------------------------------
 *
 *  Description:
 *    Processes the pos structure in the tkout and into a fiddle_j object.
 * -----------------------------------------------------------------
 */

#include "do_pos.h"

/* #################################################################
 *		#DEFINES, TYPES and GLOBAL (PRIVATE) VARIABLES
 * ################################################################# */

PRIVATE JNIEnv *env;
PRIVATE tkout_t *tk;
PRIVATE indOb_t iObj;


/* #################################################################
 *		PRIVATE functions
 * ################################################################# */

#ifdef EBUG
PRIVATE void passo(float f) {
  fprintf(stderr, "do_pos->Passo %.2f.\n", f);
  fflush(stderr);
}
#endif


/*
 * For every node in the data field of the pos structure (data_t) this
 * function is aplyed, through the iteration of Chain_walk(...). For
 * every data in the node a corresponding JAVA image is created and
 * enqueued in the data field (chain_j) of the fiddle_j pos object
 * (data_j).
 * (+) token, the data in the node (see chain_walk() );
 * (+) no_args, extra arguments. This is ignored!
 * (-) A success value. If negative the chain_walk function will stop
 * the traversing.
 */
PRIVATE int process_pos_structure(void *token, void *no_args) {
	jclass      cls;
	jmethodID   mid;
	jobject     ppos;

	pposition_t *data = (pposition_t *) token;
	jstring     file  = (*env)->NewStringUTF(env, data->file);
	jint        line  = data->line;
	jint        pos   = data->pos;
	jstring     where = (*env)->NewStringUTF(env, data->where);
	jlong       frame = (unsigned int) data->frame;

#ifdef EBUG
	fprintf(stdout, "frame as address: %#x\nframe as long: %lu\n", (unsigned int) frame, (unsigned long) frame);
#endif
	
	cls = (*env)->FindClass(env, POSITION_J_SIG);
	if(cls == 0) fatalError("process_pos:process_pos_structure:FindClass:Position_j\n", -16);
	mid = (*env)->GetMethodID(env, cls, "<init>", "(Ljava/lang/String;IILjava/lang/String;J)V");
	if(mid == 0) fatalError("process_pos:process_pos_structure:GetMethodID: Position_j Constructor\n", -2);
	ppos = (*env)->NewObject(env, cls, mid, file, line, pos, where, frame);

	insertObjInJAVA(env, ppos, iObj->pos);

	return 0;
}


/* #################################################################
 *		PUBLIC functions
 * ################################################################# */

PUBLIC void process_pos(JNIEnv *envArg, indOb_t iObjArg, tkout_t *tkArg) {
  jclass    cls;
  jmethodID mid;
  jobject   str;
  jobject   code;
  jobject   data;
  char      buffer[MAX_STR];

#ifdef EBUG
  passo(0);
#endif
  
  env = envArg;
  iObj = iObjArg;
  tk = tkArg;

  if(tkArg->pos != NULL) {
	// There's work to do, let's do it then!

#ifdef EBUG
	passo(1);
#endif
  
// The string field from the pos structure
	if(tk->pos->str == NULL) tk->pos->str=strdup("");
	str = (*env)->NewStringUTF(env, tk->pos->str);

#ifdef EBUG
	fprintf(stderr, "tk->pos->str=%s\nstr=%s\n", tk->pos->str, (*env)->GetStringUTFChars(env, str, NULL));
  passo(2);
#endif
  
// The code field from the pos structure
	cls = (*env)->FindClass(env, CODE_J_SIG);
	if(cls == 0) fatalError("process_pos:FindClass:Code_j\n", -1);
	mid = (*env)->GetMethodID(env, cls, "<init>", "(I)V");
	if(mid == 0) fatalError("process_pos:GetMethodID:Code_j Constructor\n", -2);
	code = (*env)->NewObject(env, cls, mid, tk->pos->code);

#ifdef EBUG
  passo(3);
#endif
  
// An empty field
	cls = (*env)->FindClass(env, CHAIN_J_SIG);
	if(cls == 0) fatalError("process_pos:FindClass:Chain_j\n", -13);
	mid = (*env)->GetMethodID(env, cls, "<init>", "()V");
	if(mid == 0) fatalError("process_pos:GetMethodID:Chain_j Constructor\n", -2);
	data = (*env)->NewObject(env, cls, mid);
	
	
#ifdef EBUG
  passo(4);
#endif
  
// Create the pos object with the str and code values set and an empty
// data (Chain_J) field
	cls = (*env)->FindClass(env, DATA_J_SIG);
	if(cls == 0) fatalError("process_pos:FindClass:Data_j\n", -14);
	mid = (*env)->GetMethodID(env, cls, "<init>", StrCodeChain_sigV(buffer));
	if(mid == 0) fatalError("process_pos:GetMethodID:Data_j Constructor\n", -2);
	iObj->pos = (*env)->NewObject(env, cls, mid, str, code, data);

#ifdef EBUG
  passo(5);
#endif
  
// Fill the data field of the pos Object
	if(chain_walk(tk->pos->data, process_pos_structure, NULL) != NULL)
	  fatalError("Error processing pos structure structure. Traversing not completed. Program canceled.", -1);
	
#ifdef EBUG
  passo(6);
#endif
  
  } else {
	
#ifdef EBUG
	passo(7);
#endif
  
// Create an empty pos object
	cls = (*env)->FindClass(env, DATA_J_SIG);
	if(cls == 0) fatalError("process_pos:FindClass:Data_j\n", -14);
	mid = (*env)->GetMethodID(env, cls, "<init>", "()V");
	if(mid == 0) fatalError("process_pos:GetMethodID:Data_j Constructor\n", -2);
	iObj->pos = (*env)->NewObject(env, cls, mid);

#ifdef EBUG
	passo(8);
#endif
  
  }
  
}
